import json
import time

from oic.oauth2 import Message

__author__ = 'roland'

# Message direction
INCOMING = 1
OUTGOING = 2

# standard event labels
EV_ASSERTION = 'assertion'
EV_CONDITION = 'condition'
EV_EXCEPTION = 'exception'
EV_END = 'end'
EV_EVENT = 'event'
EV_FAULT = 'fault'
EV_FUNCTION = 'function'
EV_HANDLER_RESPONSE = 'handler response'
EV_HTML_SRC = 'html src'
EV_HTTP_ARGS = 'http args'
EV_HTTP_INFO = 'http info'
EV_HTTP_REQUEST = 'http request'
EV_HTTP_RESPONSE = 'http response'
EV_HTTP_RESPONSE_HEADER = 'http response header'
EV_NOOP = 'not expected to do'
EV_OPERATION = 'phase'
EV_OP_ARGS = 'operation args'
EV_PROTOCOL_RESPONSE = 'protocol response'
EV_PROTOCOL_REQUEST = 'protocol request'
EV_REDIRECT_URL = 'redirect url'
EV_REPLY = 'reply'
EV_REQUEST = 'request'
EV_REQUEST_ARGS = 'request args'
EV_RESPONSE = 'response'
EV_RESPONSE_ARGS = 'response args'
EV_RUN = 'run'
EV_SEND = 'send'
EV_URL = 'url'
EV_JWS_HEADER = 'JWS header'
EV_JWE_HEADER = 'JWE header'
EV_ENDPOINT = 'endpoint'
EV_OBJ = 'object'
EV_OBJ_ARGS = 'object args'

class NoSuchEvent(Exception):
    pass


class Base(object):
    def __init__(self, **kwargs):
        self.op_args = kwargs

    def gather_args(self):
        return self.op_args

    def to_str(self):
        _res = self.gather_args()
        return json.dumps(_res, sort_keys=True, indent=4,
                          separators=(',', ': '))


class Operation(Base):
    def __init__(self, name, typ='', **kwargs):
        Base.__init__(self, **kwargs)
        self.name = name
        self.type = typ

    def gather_args(self):
        _res = {'name': self.name}
        if self.type:
            _res['type'] = self.type
        _res.update(Base.gather_args(self))
        return _res


class FailedOperation(Operation):
    def __init__(self, name, error, **kwargs):
        Operation.__init__(self, name, **kwargs)
        self.error = error

    def gather_args(self):
        _res = {'name': self.name, 'error': self.error}
        if self.type:
            _res['type'] = self.type
        _res.update(self.op_args)
        return _res


class HTTPRequest(Base):
    def __init__(self, endpoint, method):
        Base.__init__(self)
        self.endpoint = endpoint
        self.method = method
        self.authz = None

    def gather_args(self):
        _res = {}
        for k in ['endpoint', 'method','authz']:
            v = getattr(self, k)
            if v:
                _res[k] = v
        return _res


class HTTPResponse(object):
    def __init__(self, response):
        try:
            self.status_code = response.status_code
        except AttributeError:
            self.status_code = int(response.status.split(' ')[0])
        try:
            self.url = response.url
        except AttributeError:
            self.url = ''

        self.headers = response.headers

        # Should perhaps be more intelligent about this
        if self.status_code >= 400:
            try:
                self.text = response.text
            except AttributeError:
                self.text = response.message
        else:
            self.text = ''


class Event(object):
    def __init__(self, timestamp=0, typ='', data=None, ref='', sub='',
                 sender='', direction=0, **kwargs):
        self.timestamp = timestamp or time.time()
        self.typ = typ
        self.data = data
        try:
            self.ref = ref.lower()
        except AttributeError:
            self.ref = ref
        self.sub = sub
        self.sender = sender
        self.direction = direction
        self.kwargs = kwargs

    def __str__(self):
        return '{}:{}:{}'.format(self.timestamp, self.typ, self.data)

    def __eq__(self, other):
        if isinstance(other, Event):
            for param in ['timestamp', 'typ', 'data', 'ref', 'sub', 'sender']:
                if getattr(self, param) != getattr(other, param):
                    return False
        return True

    def older(self, other):
        if other.timestamp >= self.timestamp:
            return True
        return False

    def what_when(self):
        return '{}:{}'.format(self.timestamp, self.typ)


class Events(object):
    def __init__(self):
        self.events = []

    def store(self, typ, data, ref='', sub='', sender='', direction=0,
              **kwargs):
        index = time.time()

        typ = typ.lower()
        if typ == EV_HTTP_RESPONSE:  # only store part of the instance
            data = HTTPResponse(data)

        self.events.append(
            Event(int(index), typ, data, ref, sub, sender, direction, **kwargs))
        return index

    def by_index(self, index):
        l = [e for e in self.events if e.timestamp == index]
        if l:
            return l[0]
        else:
            raise KeyError(index)

    def by_ref(self, ref):
        try:
            lr = ref.lower()
        except AttributeError:
            lr = ref
        return [e for e in self.events if e.ref == lr]

    def by_direction(self, direction):
        return [e for e in self.events if e.direction == direction]

    def get(self, typ):
        ltyp = typ.lower()
        return [ev for ev in self.events if ev.typ == ltyp]

    def get_data(self, typ, sender=''):
        if sender:
            return [d.data for d in self.get(typ) if d.sender == sender]
        else:
            return [d.data for d in self.get(typ)]

    def get_messages(self, typ, msg_cls):
        res = []
        for m in self.get(typ):
            if m.data.__class__ == msg_cls:
                res.append(m.data)
        # return [m.data for m in self.get(typ) if isinstance(m.data, msg_cls)]
        return res

    def last(self, typ):
        res = self.get(typ)
        if len(res):
            return res[-1]
        else:
            return None

    def get_message(self, typ, msg_cls):
        l = self.get_messages(typ, msg_cls)
        if l:
            return l[-1]

        raise NoSuchEvent('{}:{}'.format(typ, msg_cls))

    def last_item(self, typ):
        l = self.get_data(typ)
        if l:
            return l[-1]

        raise NoSuchEvent(typ)

    def __len__(self):
        return len(self.events)

    def __getitem__(self, item):
        return self.get_data(item)

    def __setitem__(self, key, value):
        self.store(key, value)

    def append(self, event):
        assert isinstance(event, Event)
        self.events.append(event)

    def extend(self, events):
        for event in events:
            self.append(event)

    def __iter__(self):
        return self.events.__iter__()

    def last_of(self, types):
        l = self.events[:]
        l.reverse()
        for ev in l:
            if ev.typ in types:
                return ev.data

        return None

    def __contains__(self, event):
        ts = event.timestamp
        for ev in self.events:
            if event.timestamp == ev.timestamp:
                if event == ev:
                    return True

        return False

    def sort(self):
        self.events.sort(key=lambda event: event.timestamp)

    def to_html(self, form='table'):
        if form == 'list':
            text = ['<ul>']
            for ev in self.events:
                text.append('<li> {}'.format(ev))
            text.append('</ul>')
        else:
            text = ['<table border=1 width="600">']
            for ev in self.events:
                if isinstance(ev.data, Message):
                    _data = json.dumps(ev.data.to_dict(), sort_keys=True,
                                       indent=4, separators=(',', ': '))
                else:
                    _data = ev.data
                text.append(
                    '<tr><td>{time}</td><td>{typ}</td>'
                    '<td>{data}}</td></tr>'.format(
                        time=ev.timestamp, typ=ev.typ, data=_data))
            text.append('</table>')

        return '\n'.join(text)

    def __str__(self):
        return '\n'.join(['{}'.format(ev) for ev in self.events])

    def reset(self):
        self.events = []

    def when(self, typ, msg):
        res = []
        for m in self.get(typ):
            if m.data.__class__ == msg:
                res.append(m.timestamp)

        return res

    def last_event_type(self):
        return self.events[-1].typ

    def timeline(self):
        start = self.events[0].timestamp
        return [(ev.timestamp - start, ev.typ, ev.data) for ev in self.events]

    def digest(self):
        res = []
        for event in self.events:
            res.append(event.what_when())
        return res


def funtion_to_str(event):
    _dat = event.data
    res = [_dat['name']]
    if 'args' in _dat and _dat['args']:
        res.append('args:{}'.format(_dat['args']))
    if 'kwargs' in _dat and _dat['kwargs']:
        res.append('kwargs:{}'.format(_dat['kwargs']))
    return res


def http_response_to_str(event):
    _dat = event.data
    res = [event.typ]
    if _dat.url:
        res.append('url:{}'.format(_dat.url))
    if _dat.status_code:
        res.append('status_code:{}'.format(_dat.status_code))
    if _dat.text:
        res.append('message:{}'.format(_dat.text))
    return res


def exception_to_str(event):
    res = [event.typ]
    try:
        res.append('{} {}'.format(event.kwargs['note'], event.data))
    except KeyError:
        res.append('{}'.format(event.data))
    return res


def message_to_str(event):
    return [event.data.__class__.__name__,
            json.dumps(event.data.to_dict(), sort_keys=True, indent=4,
                       separators=(',', ': '))]


TO_STR = {
    EV_FUNCTION: funtion_to_str,
    EV_HTTP_RESPONSE: http_response_to_str,
    EV_EXCEPTION: exception_to_str,
    EV_PROTOCOL_REQUEST: message_to_str,
    EV_PROTOCOL_RESPONSE: message_to_str
}


def layout(start, event):
    elem = ['{}'.format(round(event.timestamp - start, 3))]
    if event.direction:
        if event.direction == OUTGOING:
            elem.append('-->')
        else:
            elem.append('<--')

    try:
        elem.extend(TO_STR[event.typ](event))
    except KeyError:
        elem.append(event.typ)
        if isinstance(event.data, Base):
            elem.append(event.data.to_str())
        else:
            elem.append(str(event.data))

    return ' '.join(elem)


def row(start, event):
    if event.typ == EV_OPERATION:
        _row = ['<tr class="info"><td>{}</td>'.format(
            round(event.timestamp - start, 3))]
    else:
        _row = ['<tr><td>{}</td>'.format(
            round(event.timestamp - start, 3))]

    try:
        p = TO_STR[event.typ](event)
        _row.append(
            '<td>{}</td><td><pre><code>{}</pre></code></td>'.format(
                p[0],' '.join(p[1:])))
    except KeyError:
        _row.append('<td style="max-width: 500px; word-wrap: break-word;">{}</td>'.format(event.typ))
        if isinstance(event.data, Base):
            _row.append('<td style="max-width: 500px; word-wrap: break-word;">{}</td>'.format(event.data.to_str()))
        else:
            _row.append('<td style="max-width: 500px; word-wrap: break-word;">{}</td>'.format(str(event.data)))
    _row.append("</tr>")
    return "".join(_row)


from future.backports import ceil
from math import log

__author__ = 'roland'


def hist(source):
    hist = {}
    l = 0
    for e in source:
        l += 1
        if e not in hist:
            hist[e] = 0
        hist[e] += 1
    return (l, hist)


def entropy(hist, l):
    elist = []
    for v in hist.values():
        c = v / l
        elist.append(-c * log(c, 2))
    return sum(elist)


def calculate(source):
    (l, h) = hist(source)
    return l * ceil(entropy(h, l))


if __name__ == '__main__':
    print(calculate('InAGkrHTT7Yz1gDk'))


# -*- coding: utf-8 -*-

"""
JSON 2 HTML convertor
=====================

(c) Varun Malhotra 2013
Source Code: https://github.com/softvar/json2html

Rewritten to work with Python3

Contributors:
-------------
1. Michel Müller(@muellermichel), https://github.com/softvar/json2html/pull/2
2. Roland Hedberg

LICENSE: MIT
--------
"""


from past.builtins.misc import unicode
import json


class Json2Html(object):
    def __init__(self):
        # table attributes such as class, id, data-attr-*, etc.
        # eg:
        # table_attributes = 'class = "table table-bordered sortable"'
        self.root_table_attributes = ''
        self.table_attributes = ''

    def convert(self, json_input='', **kwargs):
        """
            Convert JSON to HTML Table format
        """

        if 'root_table_attributes' in kwargs:
            self.root_table_attributes = kwargs['root_table_attributes']
        else:
            # by default HTML table border
            self.root_table_attributes = 'border="1"'

        if 'table_attributes' in kwargs:
            self.table_attributes = kwargs['table_attributes']
        else:
            # by default HTML table border
            self.table_attributes = ''

        if json_input:
            try:
                json.loads(json_input)
            except:
                json_input = json.dumps(json_input)
        else:
            raise Exception('Can\'t convert NULL!')

        input = json.loads(json_input)
        return self.iter_json(input, root=True)

    @staticmethod
    def column_headers_from_list_of_dicts(input):
        """
            If suppose some key has array of objects and all the keys are same,
            instead of creating a new row for each such entry,
            club such values, thus it makes more sense and more readable table.

            @example:
                jsonObject = {
                    "sampleData": [
                        {"a":1, "b":2, "c":3},
                        {"a":5, "b":6, "c":7}
                    ]
                }
                OUTPUT:
                _____________________________
                |               |   |   |   |
                |               | a | c | b |
                |   sampleData  |---|---|---|
                |               | 1 | 3 | 2 |
                |               | 5 | 7 | 6 |
                -----------------------------

            @contributed by: @muellermichel
        """

        if len(input) < 2:
            return None
        if not isinstance(input[0], dict):
            return None

        column_headers = input[0].keys()

        for entry in input:
            if not isinstance(entry, dict):
                return None
            if len(list(entry.keys())) != len(column_headers):
                return None
            for header in column_headers:
                if header not in entry:
                    return None
        return column_headers

    def iter_json(self, input, root=False):
        """
            Iterate over the JSON and process it
            to generate the super awesome HTML Table format
        """

        def markup(entry):
            """
                Check for each value corresponding to its key
                and return accordingly
            """
            if isinstance(entry, unicode):
                return unicode(entry)
            if isinstance(entry, int) or isinstance(entry, float):
                return str(entry)
            if isinstance(entry, list) and len(entry) == 0:
                return ''
            if isinstance(entry, list):
                listMarkup = ''
                listMarkup += '<ul><li>'
                listMarkup += '</li><li>'.join(
                    [markup(child) for child in entry])
                listMarkup += '</li></ul>'
                return listMarkup
            if isinstance(entry, dict):
                return self.iter_json(entry)

            # safety: don't do recursion over anything that we don't know about
            return ''

        converted_output = ''

        if root:
            table_init_markup = "<table {}>".format(self.root_table_attributes)
        else:
            table_init_markup = "<table {}>".format(self.table_attributes)

        converted_output += table_init_markup

        try:
            for (k, v) in input.items():
                converted_output += '<tr>'
                converted_output = converted_output + '<th>' + markup(k) + '</th>'

                if v is None:
                    v = unicode("")
                if isinstance(v, list):
                    column_headers = self.column_headers_from_list_of_dicts(v)
                    if column_headers is not None:
                        converted_output += '<td>'
                        converted_output += table_init_markup
                        converted_output += '<tr><td>' + '</td><td>'.join(
                            column_headers) + '</td></tr>'
                        for list_entry in v:
                            converted_output = converted_output
                            converted_output += '<tr><td>'
                            converted_output += '</td><td>'.join(
                                [markup(list_entry[column_header]) for
                                 column_header in column_headers])
                            converted_output += '</td></tr>'

                        converted_output += '</table></td></tr>'
                        continue
                converted_output = converted_output + '<td>' + markup(v) + '</td>'
                converted_output += '</tr>'
            converted_output += '</table>'

        except:
            raise Exception('Not a valid JSON list')
        return converted_output


json2html = Json2Html()



import logging
# from urllib.parse import parse_qs
from urllib.parse import quote_plus

import cherrypy
from oic.utils.http_util import Redirect
from oic.utils.http_util import Response

from otest import Break, ConditionError
from otest import Done
from otest import exception_trace
from otest import tool
from otest.check import CRITICAL
from otest.check import ERROR
from otest.check import OK
from otest.check import State
from otest.events import EV_CONDITION, EV_OPERATION
from otest.events import EV_REDIRECT_URL
from otest.prof_util import compress_profile, from_profile, to_profile
from otest.result import Result
from otest.result import safe_path
from otest.verify import Verify

__author__ = 'roland'

logger = logging.getLogger(__name__)


class Tester(tool.Tester):
    def __init__(self, io, sh, profiles, flows=None,
                 check_factory=None, msg_factory=None, cache=None,
                 map_prof=None, **kwargs):
        tool.Tester.__init__(self, io, sh, flows=flows,
                             msg_factory=msg_factory, cache=cache,
                             check_factory=check_factory, **kwargs)
        self.profiles = profiles
        self.map_prof = map_prof
        self.conv = None

    def match_profile(self, test_id):
        _spec = self.flows[test_id]
        return self.map_prof(self.sh.profile.split("."),
                             _spec["profile"].split("."))

    def fname(self, test_id):
        try:
            return safe_path(
                self.conv.entity.provider_info['issuer'],
                self.sh.profile, test_id)
        except KeyError:
            return safe_path('dummy', self.sh.profile, test_id)

    def run_flow(self, test_id, index=0, profiles=None, conf=None):
        logger.info("<=<=<=<=< %s >=>=>=>=>" % test_id)
        self.flows.complete[test_id] = False
        self.conv.test_id = test_id
        self.conv.conf = conf

        if index >= len(self.conv.sequence):
            return CRITICAL

        res = Result(self.sh, self.kwargs['profile_handler'])

        _oper = None
        for item in self.conv.sequence[index:]:
            self.sh["index"] = index
            if isinstance(item, tuple):
                cls, funcs = item
            else:
                cls = item
                funcs = {}

            _name = cls.__name__
            _line = "<--<-- {} --- {} -->-->".format(index, _name)
            logger.info(_line)
            self.conv.events.store(EV_OPERATION, _line)
            try:
                _oper = cls(conv=self.conv, inut=self.inut, sh=self.sh,
                            profile=self.sh.profile, test_id=test_id, conf=conf,
                            funcs=funcs, check_factory=self.check_factory,
                            cache=self.cache,
                            tool_conf=self.kwargs['tool_conf'])
                self.conv.operation = _oper
                _oper.setup(self.profiles.PROFILEMAP)
                if _oper.fail:
                    break
                resp = _oper()
            except Break:
                break
            except cherrypy.HTTPError:
                raise
            except Exception as err:
                self.conv.events.store(
                    EV_CONDITION,
                    State(test_id=test_id, status=ERROR, message=err,
                          context=cls.__name__))
                _trace = exception_trace(cls.__name__, err, logger)
                self.sh["index"] = index
                self.store_result(res)
                return {'exception_trace': _trace}
            else:
                rsp = self.handle_response(resp, index)
                if rsp:
                    self.store_result(res)
                    return self.inut.respond(rsp)

            index += 1
            if index == len(self.conv.sequence):
                break

        if isinstance(_oper, Done):
            try:
                if self.conv.flow["assert"]:
                    _ver = Verify(self.check_factory, self.conv)
                    _ver.test_sequence(self.conv.flow["assert"])
            except (KeyError, Break):
                self.conv.events.store(EV_CONDITION, State('Done', status=OK))
            except ConditionError:
                pass
            except Exception as err:
                raise
            else:
                self.conv.events.store(EV_CONDITION, State('Done', status=OK))

        tinfo = self.store_result(res)
        return tinfo['state']


class ClTester(Tester):
    pass


class WebTester(Tester):
    def display_test_list(self):
        try:
            if self.sh.session_init():
                return self.inut.flow_list()
            else:
                try:
                    resp = Redirect("%sopresult#%s" % (
                        self.kwargs['base_url'], self.sh["testid"][0]))
                except KeyError:
                    return self.inut.flow_list()
                else:
                    return resp(self.inut.environ, self.inut.start_response)
        except Exception as err:
            exception_trace("display_test_list", err)
            return self.inut.err_response("session_setup", err)

    def set_profile(self, info):
        try:
            old = from_profile(self.sh.profile)

            new = from_profile(to_profile(info))
            for attr in ['enc', 'extra', 'none', 'return_type', 'sig', 'form_post']:
                old[attr] = new[attr]

            # Store new configuration
            try:
                rest = self.sh.extra['rest']
            except KeyError:
                self.conv.tool_conf.update(compress_profile(old))
            else:
                qp = [quote_plus(p) for p in [self.sh.iss, self.sh.tag]]
                _, _conf = rest.read_conf(*qp)
                _conf['tool'].update(compress_profile(old))
                rest.store(qp[0], qp[1], _conf)

                # This will fail if no test has been run before the conf
                # is changed
                try:
                    self.conv.tool_conf = _conf['tool']
                except AttributeError:
                    pass

            # reset all test flows
            self.flows.test_info = {}
            self.flows.complete = {}
            self.sh.reset_session(profile=old['profile'])
            # Back to test list
            return self.inut.flow_list()
        except Exception as err:
            return self.inut.err_response("profile", err)

    def handle_response(self, resp, index, oper=None):
        if resp:
            self.sh["index"] = index
            if isinstance(resp, Response):
                if self.conv.events.last_event_type() != EV_REDIRECT_URL:
                    self.conv.events.store(EV_REDIRECT_URL, resp.message)
                return resp(self.inut.environ, self.inut.start_response)
            else:
                return resp
        else:
            return None

    def cont(self, **kwargs):
        path = kwargs['path']
        index = int(kwargs['index'])
        index += 1

        self.store_result()

        try:
            return self.run_flow(path, index=index)
        except cherrypy.HTTPRedirect:
            raise
        except Exception as err:
            exception_trace("", err, logger)
            self.store_result()
            return self.inut.err_response("run", err)

    def async_response(self, conf, response=None):
        index = self.sh["index"]
        item = self.sh["sequence"][index]
        self.conv = self.sh["conv"]

        if isinstance(item, tuple):
            cls, funcs = item
        else:
            cls = item

        logger.info("<--<-- {} --- {}".format(index, cls.__name__))
        resp = self.conv.operation.parse_response(self.sh["testid"],
                                                  self.inut,
                                                  self.message_factory,
                                                  response=response)
        if resp:
            return resp

        index += 1

        return self.run_flow(self.sh["testid"], index=index)

    def handle_request(self, request=None, request_args=None):
        index = self.sh["index"]
        item = self.sh["sequence"][index]
        self.conv = self.sh["conv"]

        if isinstance(item, tuple):
            cls, funcs = item
        else:
            cls = item
            funcs = {}

        _line = "<--<-- {} --- {} -->-->".format(index, cls.__name__)
        logger.info(_line)
        self.conv.events.store(EV_OPERATION, _line)
        _oper = cls(conv=self.conv, inut=self.inut, sh=self.sh,
                    profile=self.sh.profile, test_id=self.sh["testid"],
                    conf=self.conv.conf, funcs=funcs,
                    check_factory=self.check_factory, cache=self.cache,
                    tool_conf=self.kwargs['tool_conf'])
        self.conv.operation = _oper
        resp = self.conv.operation.handle_request(self.message_factory,
                                                  request=request,
                                                  request_args=request_args)

        if resp:
            return resp

        index += 1
        return self.run_flow(self.sh["testid"], index=index)



import logging

from oic.utils.http_util import Response, NotFound

from otest.summation import represent_result
from otest.check import ERROR
from otest.check import OK
from otest.check import WARNING
from otest.check import INCOMPLETE
from otest.handling import InfoHandling

__author__ = 'roland'

logger = logging.getLogger(__name__)

TEST_RESULTS = {OK: "OK", ERROR: "ERROR", WARNING: "WARNING",
                INCOMPLETE: "INCOMPLETE"}

def get_test_info(session, test_id):
    return session["test_info"][test_id]


class WebIh(InfoHandling):
    def __init__(self, conf, flows, profile_handler, profile='', lookup=None,
                 cache=None, environ=None, start_response=None, session=None,
                 **kwargs):
        InfoHandling.__init__(self, flows, profile, profile_handler,
                              cache, session=session, **kwargs)

        self.conf = conf
        self.lookup = lookup
        self.environ = environ
        self.start_response = start_response

        try:
            self.base_url = kwargs['base_url']
        except KeyError:
            try:
                self.base_url = kwargs['base']
            except KeyError:
                self.base_url = conf.BASE

    def flow_list(self, **kwargs):
        resp = Response(mako_template="flowlist.mako",
                        template_lookup=self.lookup,
                        headers=[])

        argv = {
            "flows": self.flows.display_info(self.session['tests']),
            "profile": self.session["profile"],
            #"test_info": list(self.session["test_info"].keys()),
            "base": self.base_url,
        }

        if not argv['base'].endswith('/'):
            argv['base'] += '/'

        argv.update(kwargs)
        return resp(self.environ, self.start_response, **argv)

    def test_info(self, testid):
        resp = Response(mako_template="testinfo.mako",
                        template_lookup=self.lookup,
                        headers=[])

        info = self.flows.test_info[testid]

        argv = {
            "profile": info["profile_info"],
            "events": info["events"],
            "result": info['result'],
            "base": self.base_url,
        }

        return resp(self.environ, self.start_response, **argv)

    def not_found(self):
        """Called if no URL matches."""
        resp = NotFound()
        return resp(self.environ, self.start_response)



import inspect
import json
import logging
import sys

from oic.oauth2 import SUCCESSFUL
from oic.oauth2.message import ErrorResponse
from oic.oauth2.message import MissingRequiredAttribute
from oic.oauth2.message import AuthorizationResponse

from otest.check import Check, Warnings
from otest.check import CRITICAL
from otest.check import CriticalError
from otest.check import ERROR
from otest.check import Error
from otest.check import ExpectedError
from otest.check import get_protocol_response
from otest.check import WARNING
from otest.events import EV_HTTP_RESPONSE
from otest.events import EV_PROTOCOL_RESPONSE
from otest.events import EV_RESPONSE
from otest.events import NoSuchEvent

logger = logging.getLogger(__name__)

__author__ = 'rolandh'

CONT_JSON = "application/json"
CONT_JWT = "application/jwt"


def _last_response(conv, typ, error):
    try:
        return conv.events.last_item(typ)
    except NoSuchEvent:
        logger.warning(error)
        logger.warning('\n'.join(conv.events.digest()))
        raise


def last_http_response(conv):
    return _last_response(conv, EV_HTTP_RESPONSE, 'No HTTP Response ??')


def last_raw_response(conv):
    return _last_response(conv, EV_RESPONSE, 'No Response ??')


def last_protocol_response(conv):
    return _last_response(conv, EV_PROTOCOL_RESPONSE, 'No parsed Response ??')


class MissingRedirect(CriticalError):
    """ At this point in the flow a redirect back to the client was expected.
    """
    cid = "missing-redirect"
    msg = "Expected redirect to the RP, got something else"

    def _func(self, conv=None):
        self._status = self.status
        return {"url": conv.position}


class Parse(CriticalError):
    """ Parsing the response """
    cid = "response-parse"
    errmsg = "Parse error"

    def _func(self, conv=None):
        if conv.exception:
            self._status = self.status
            err = conv.exception
            self._message = "%s: %s" % (err.__class__.__name__, err)
        else:
            _rmsg = conv.response_message
            cname = _rmsg.type()
            if conv.cresp.response != cname:
                self._status = self.status
                self._message = (
                    "Didn't get a response of the type expected:",
                    " '%s' instead of '%s', content:'%s'" % (
                        cname, conv.response_type, _rmsg))
                return {
                    "response_type": conv.response_type,
                    "url": conv.position
                }

        return {}


class CheckHTTPResponse(CriticalError):
    """
    Checks that the HTTP response status is within the 200 or 300 range.
    Also does some extra JSON checks
    """
    cid = "check-http-response"
    msg = "OP error"

    def _func(self, conv):
        res = {}
        try:
            _response = last_http_response(conv)
        except NoSuchEvent:
            return res

        if _response.status_code >= 400:
            self._status = self.status
            self._message = self.msg
            if CONT_JSON in _response.headers["content-type"]:
                try:
                    err = ErrorResponse().deserialize(_response.txt, "json")
                    self._message = err.to_json()
                except Exception:
                    res["content"] = _response.text
            else:
                res["content"] = _response.text
            res["url"] = _response.url
            res["http_status"] = _response.status_code
        elif _response.status_code in [300, 301, 302]:
            pass
        else:
            # might still be an error message
            msg = conv.events.last_item(EV_PROTOCOL_RESPONSE)
            if isinstance(msg, ErrorResponse):
                self._message = msg.to_json()
                self._status = self.status

        return res


class CheckHTTPErrorResponse(Warnings):
    """
    Checks that an error code is either 400 or 401 which are the only ones
    accepted by OAuth2/OIDC.
    """
    cid = "check-http-error-response"
    msg = "OP error"

    def _func(self, conv):
        res = {}

        try:
            _response = last_http_response(conv)
        except NoSuchEvent:
            return res

        if _response.status_code not in [200, 300, 301, 302, 400, 401]:
            self._status = self.status
            self._message = self.msg
            if CONT_JSON in _response.headers["content-type"]:
                try:
                    err = ErrorResponse().deserialize(_response.txt, "json")
                    self._message = err.to_json()
                except Exception:
                    res["content"] = _response.text
            else:
                res["content"] = _response.text
            res["url"] = _response.url
            res["http_status"] = _response.status_code

        return res


class VerifyErrorResponse(ExpectedError):
    """
    Verifies that the response received by the client via redirect was an Error
    response.
    """
    cid = "verify-err-response"
    msg = "OP error"

    def _func(self, conv):
        res = {}

        try:
            response = last_http_response(conv)
        except NoSuchEvent:
            return res

        if response.status_code == 302:
            _loc = response.headers["location"]
            if "?" in _loc:
                _query = _loc.split("?")[1]
            elif "#" in _loc:
                _query = _loc.split("#")[1]
            else:
                self._message = "Faulty error message"
                self._status = ERROR
                return

            try:
                err = ErrorResponse().deserialize(_query, "urlencoded")
                err.verify()
                # res["temp"] = err
                res["message"] = err.to_dict()
            except Exception:
                self._message = "Faulty error message"
                self._status = ERROR
        else:
            self._message = "Expected a redirect with an error message"
            self._status = ERROR

        return res


class CheckRedirectErrorResponse(ExpectedError):
    """
    Checks that the HTTP response status is outside the 200 or 300 range
    or that an error message has been received urlencoded in the form of a
    redirection.
    """
    cid = "check-redirect-error-response"
    msg = "OP error"

    def _func(self, conv):
        res = {}

        try:
            _response = last_http_response(conv)
        except NoSuchEvent:
            return res

        try:
            _loc = _response.headers["location"]
            if "?" in _loc:
                query = _loc.split("?")[1]
            elif "#" in _loc:
                query = _loc.split("#")[1]
            else:  # ???
                self._message = "Expected a redirect"
                self._status = CRITICAL
                return res
            env_id = conv.events.store(EV_RESPONSE, query)
        except (KeyError, AttributeError):
            self._message = "Expected a redirect"
            self._status = CRITICAL
            return res

        if _response.status_code == 302:
            err = ErrorResponse().deserialize(query, "urlencoded")
            try:
                err.verify()
                res["content"] = err.to_json()
                conv.events.store(EV_PROTOCOL_RESPONSE, err, env_id)
            except MissingRequiredAttribute:
                self._message = "Expected an error message"
                self._status = CRITICAL
        else:
            self._message = "Expected an error message"
            self._status = CRITICAL

        return res


class VerifyBadRequestResponse(ExpectedError):
    """
    Verifies that the OP returned a 400 Bad Request response containing a
    Error message.
    """
    cid = "verify-bad-request-response"
    msg = "OP error"

    def _func(self, conv):
        res = {}

        try:
            _response = last_http_response(conv)
        except NoSuchEvent:
            return res

        if _response.status_code == 400:
            err = ErrorResponse().deserialize(_response.text, "json")
            try:
                err.verify()
            except MissingRequiredAttribute:
                try:
                    self._status = self._kwargs["status"]
                except KeyError:
                    self._status = ERROR
                self._message = "Expected an error message"
            else:
                res["content"] = err.to_json()
        elif _response.status_code in [301, 302, 303]:
            pass
        elif _response.status_code < 300:
            _content = conv.events.last_item(EV_RESPONSE)
            err = ErrorResponse().deserialize(_content, "json")
            try:
                err.verify()
            except MissingRequiredAttribute:
                try:
                    self._status = self._kwargs["status"]
                except KeyError:
                    self._status = ERROR
                self._message = "Expected an error message"
            else:
                res["content"] = err.to_json()
            conv.events.store(EV_PROTOCOL_RESPONSE, err)
        else:
            self._message = "Expected an error message"
            try:
                self._status = self._kwargs["status"]
            except KeyError:
                self._status = CRITICAL

        return res


class VerifyRandomRequestResponse(ExpectedError):
    cid = "verify-random-request-response"
    msg = "OP error"

    def _func(self, conv):
        res = {}

        try:
            _response = last_http_response(conv)
        except NoSuchEvent:
            return res

        try:
            _content = last_raw_response(conv)
        except NoSuchEvent:
            return res

        if _response.status_code == 400:
            err = ErrorResponse().deserialize(_content, "json")
            err.verify()
            res["content"] = err.to_json()
            conv.events.store(EV_PROTOCOL_RESPONSE, err)
            pass
        elif _response.status_code in [301, 302, 303]:
            pass
        elif _response.status_code in SUCCESSFUL:
            err = ErrorResponse().deserialize(_content, "json")
            err.verify()
            res["content"] = err.to_json()
            conv.events.store(EV_PROTOCOL_RESPONSE, err)
        else:
            self._message = "Expected a 400 error message"
            self._status = CRITICAL

        return res


class VerifyUnknownClientIdResponse(ExpectedError):
    cid = "verify-unknown-client-id-response"
    msg = "OP error"

    def _func(self, conv):
        res = {}
        try:
            _response = last_http_response(conv)
        except NoSuchEvent:
            return res

        try:
            _content = last_raw_response(conv)
        except NoSuchEvent:
            return res

        if _response.status_code == 400:
            err = ErrorResponse().deserialize(_content, "json")
            err.verify()
            res["content"] = err.to_json()
            conv.events.store(EV_PROTOCOL_RESPONSE, err)
        elif _response.status_code in [301, 302, 303]:
            pass
        elif _response.status_code in SUCCESSFUL:
            err = ErrorResponse().deserialize(_content, "json")
            err.verify()
            res["content"] = err.to_json()
            conv.events.store(EV_PROTOCOL_RESPONSE, err)
        else:
            self._message = "Expected a 400 error message"
            self._status = CRITICAL

        return res


class VerifyError(Error):
    """
    Verifies that an error message was returned and also if it's the correct
    type.
    """
    cid = "verify-error"

    def _func(self, conv):
        try:
            response = last_http_response(conv)
        except NoSuchEvent:
            return {}

        if response.status_code == 400:
            try:
                item = json.loads(response.text)
            except Exception:
                self._message = "Expected an error response"
                self._status = self.status
                return {}
        else:
            try:
                item = conv.events.last_item(EV_PROTOCOL_RESPONSE)
            except IndexError:
                self._message = "Expected a message"
                self._status = CRITICAL
                return {}

            try:
                assert item.type().endswith("ErrorResponse")
            except AssertionError:
                self._message = "Expected an error response"
                self._status = self.status
                return {}

        try:
            assert item["error"] in self._kwargs["error"]
        except AssertionError:
            self._message = "Wrong type of error, got %s" % item["error"]
            self._status = WARNING

        return {}


class CheckErrorResponse(ExpectedError):
    """
    Checks that the HTTP response status is outside the 200 or 300 range
    or that an JSON encoded error message has been received
    """
    cid = "check-error-response"
    msg = "OP error"

    def _func(self, conv):
        res = {}
        # did I get one, should only be one
        try:
            _ = get_protocol_response(conv, ErrorResponse)[0]
        except ValueError:
            pass
        else:
            return res

        try:
            _response = last_http_response(conv)
        except NoSuchEvent:
            return {}

        if _response.status_code >= 400:
            content_type = _response.headers["content-type"]
            _content = _response.text
            if content_type is None:
                res["content"] = _content
            elif CONT_JSON in content_type:
                try:
                    self.err = ErrorResponse().deserialize(_content, "json")
                    self.err.verify()
                    res["content"] = self.err.to_json()
                except Exception:
                    res["content"] = _content
            else:
                res["content"] = _content
        elif _response.status_code in [300, 301, 302, 303]:
            pass
        else:
            # Should never get here
            self._message = 'Not an error message ??'
            self._status = WARNING

        return res


class VerifyErrorMessage(ExpectedError):
    """
    Checks that the last response was a JSON encoded error message
    """
    cid = "verify-error-response"
    msg = "OP error"

    def _func(self, conv):
        try:
            inst = last_protocol_response(conv)
        except NoSuchEvent:
            return {}

        try:
            assert isinstance(inst, ErrorResponse)
        except AssertionError:
            self._message = "Expected an error message"
            try:
                self._status = self._kwargs["status"]
            except KeyError:
                self._status = ERROR
        else:
            try:
                assert inst["error"] in self._kwargs["error"]
            except AssertionError:
                self._message = "Unexpected error type: %s" % inst["error"]
                self._status = WARNING
            except KeyError:
                pass

        return {}


class VerifyAuthnResponse(ExpectedError):
    """
    Checks that the last response was a JSON encoded authentication message
    """
    cid = "verify-authn-response"
    msg = "OP error"

    def _func(self, conv):
        try:
            inst = last_protocol_response(conv)
        except NoSuchEvent:
            return {}

        try:
            assert isinstance(inst, AuthorizationResponse)
        except AssertionError:
            self._message = "Expected an authorization response"
            self._status = ERROR

        return {}


class VerifyAuthnOrErrorResponse(ExpectedError):
    """
    Checks that the last response was a JSON encoded authentication or
    error message
    """
    cid = "authn-response-or-error"
    msg = "Expected authentication response or error message"

    def _func(self, conv):
        try:
            inst = last_protocol_response(conv)
        except NoSuchEvent:
            return {}

        try:
            assert isinstance(inst, AuthorizationResponse)
        except AssertionError:
            try:
                assert isinstance(inst, ErrorResponse)
            except AssertionError:
                self._message = "Expected an authorization or error response"
                self._status = ERROR
            else:
                try:
                    assert inst["error"] in self._kwargs["error"]
                except AssertionError:
                    self._message = "Unexpected error response: %s" % inst[
                        "error"]
                    self._status = WARNING
                except KeyError:
                    pass

        return {}


class VerifyResponse(ExpectedError):
    """
    Checks that the last response was one of a possible set of OpenID Connect
    Responses
    """
    cid = "verify-response"
    msg = "Expected OpenID Connect response"
    doc = """
    :param response_cls: Which responses the test tool has received
    :type response_cls: list of strings

    Example:
        "verify-response": {
          "response_cls": [
            "AuthorizationResponse",
            "AccessTokenResponse"
          ]
        }
    """

    def _func(self, conv):
        try:
            inst = last_protocol_response(conv)
        except NoSuchEvent:
            return {}

        try:
            _status = self._kwargs["status"]
        except KeyError:
            _status = ERROR

        if inst.__class__.__name__ not in self._kwargs["response_cls"]:
            if isinstance(inst, ErrorResponse):
                try:
                    assert inst["error"] in self._kwargs["error"]
                except AssertionError:
                    self._message = "Unexpected error response: %s" % inst[
                        "error"]
                    self._status = WARNING
                    return {}
                except KeyError:
                    pass
            else:
                self._message = "Got a %s response !?" % inst.__class__.__name__
                self._status = _status

        return {}


def factory(cid):
    for name, obj in inspect.getmembers(sys.modules[__name__]):
        if inspect.isclass(obj) and issubclass(obj, Check):
            try:
                if obj.cid == cid:
                    return obj
            except AttributeError:
                pass

    from otest.check import factory as a_factory

    return a_factory(cid)



import logging
import os

from future.backports.urllib.parse import quote
from oic.oauth2.provider import Provider

from otest.check import CRITICAL
from otest.check import ERROR
from otest.check import NOT_APPLICABLE
from otest.check import OK
from otest.check import WARNING
from otest.check import INCOMPLETE
from otest.summation import condition
from otest.summation import represent_result
from otest.summation import result_code
from otest.summation import trace_output
from otest.time_util import in_a_while

SIGN = {OK: "+", WARNING: "!", ERROR: "-", INCOMPLETE: "?",
        NOT_APPLICABLE: 'N/A', CRITICAL: "X"}

TEST_RESULTS = {OK: "OK", ERROR: "ERROR", WARNING: "WARNING",
                INCOMPLETE: "INCOMPLETE"}

logger = logging.getLogger(__name__)


def get_issuer(conv):
    try:
        return conv.info['issuer']  # dynamically acquired using WebFinger
    except KeyError:
        try:
            # From provider info discovery, dynamic or static
            return conv.entity.provider_info['issuer']
        except (KeyError, AttributeError):
            try:
                # Initial configuration
                return conv.tool_config['issuer']
            except KeyError:
                return 'unknown'


def safe_url(url):
    if url.startswith('https://'):
        url = 's_' + url[8:]
    elif url.startswith('http://'):
        url = url[7:]

    s = quote(url)
    s = s.replace('/', '%2F')
    s = s.replace('%', '')
    s = s.replace('//', '/')
    return s


def safe_path(eid, ext='', *args):
    """

    :param eid: Entity (Issuer) ID, a URL
    :param ext: Log file extension
    :param args: Additional arguments
    :return: A URL and reverse proxy safe path
    """
    s = safe_url(eid)
    path = 'log/{}'.format(s)
    for arg in args[:-1]:
        path = '{}/{}'.format(path, arg)

    if not os.path.isdir(path):
        os.makedirs(path)

    fname = '{}/{}'.format(path, args[-1])
    if ext:
        fname += '.{}'.format(ext)
    return fname


class Result(object):
    """
    Reads and writes test result information to files on disc.
    Keeps a cache for quick access.
    """

    def __init__(self, session, profile_handler):
        self.profile_handler = profile_handler
        self.session = session
        self.cache = {}
        self.logfile_extension = 'txt'

    def print_result(self, events):
        return represent_result(events)

    def op_based(self, test_id, tag=''):
        _sess = self.session
        _iss = _sess.iss
        if _iss.endswith('/' + test_id):
            _iss = _iss[:-(len(test_id) + 1)]
        if not tag:
            tag = _sess.tag
        return safe_path(_iss, self.logfile_extension, tag, _sess.profile,
                         test_id)

    def rp_based(self, test_id, tag=''):
        _sess = self.session
        return safe_path(_sess['test_conf']['start_page'],
                         self.logfile_extension, _sess.profile, test_id)

    def write_info(self, tinfo, test_id='', file_name=None, tag=''):
        if not test_id:
            test_id = tinfo['test_id']

        if file_name is None:
            if isinstance(self.session['conv'].entity, Provider):
                file_name = self.rp_based(test_id, tag)
            else:
                file_name = self.op_based(test_id, tag)

        if 'conv' not in self.session:
            return
        else:
            _conv = self.session["conv"]

        sline = 60 * "="

        _pi = tinfo['profile_info']

        output = ["Test tool version: {}".format(self.session.tool_version)]
        if _pi:
            _keys = list(_pi.keys())
            _keys.sort()
            output.extend(["%s: %s" % (k, _pi[k]) for k in _keys])
        else:
            output.extend(['Test ID: {}'.format(_conv.test_id),
                           "Timestamp: {}".format(in_a_while())])

        _events = tinfo["events"]
        output.extend(["", sline, ""])
        output.extend(trace_output(_events))
        output.extend(["", sline, ""])
        output.extend(condition(_events))
        output.extend(["", sline, ""])
        # and lastly the result
        output.append("RESULT: {}".format(self.print_result(_events)))
        output.append("")

        txt = "\n".join(output)

        f = self._open_file(file_name, 'w')
        f.write(txt)
        f.close()

        self.cache[test_id] = {'result': result_code(_events),
                               'file_name': file_name}

    def _open_file(self, file_name, mode='w'):
        try:
            fp = open(file_name, mode)
        except IOError:
            try:
                os.makedirs(os.path.dirname(file_name))
            except OSError:
                pass

            try:
                fp = open(file_name, mode)
            except Exception as err:
                logger.error(
                    "Couldn't dump to log file {} reason: {}").format(
                    file_name, err)
                raise
        return fp

    def test_status(self, tid):
        return self.cache[tid]



import datetime
import psutil

def find_test_instance(iss, tag):
    match = {"-i": iss, "-t": tag}

    for proc in psutil.process_iter():
        try:
            cmd = proc.cmdline()
        except psutil.AccessDenied:
            continue

        if len(cmd) < 5:
            continue

        flag = 0
        for first, second in match.items():
            try:
                i = cmd.index(first)
            except ValueError:
                break

            if cmd[i + 1] != second:
                break
            else:
                flag += 1

        if flag == len(match):
            return proc
    return None


def find_test_instances(prog):
    pid = {}
    for proc in psutil.process_iter():
        _name = proc.name()
        if _name in ['python', 'python3', 'Python', 'Python3']:
            try:
                cmd = proc.cmdline()
            except psutil.AccessDenied:
                continue

            if len(cmd) > 5:
                if cmd[1].endswith(prog):
                    i = cmd.index('-i')
                    iss = cmd[i + 1]
                    i = cmd.index('-t')
                    tag = cmd[i + 1]
                    i = cmd.index('-p')
                    port = cmd[i + 1]
                    since = datetime.datetime.fromtimestamp(
                        proc.create_time()).strftime(
                        "%Y-%m-%d %H:%M:%S")
                    pid[proc.pid] = {'iss': iss, 'tag': tag, 'port': port,
                                     'since': since}
    return pid


def kill_test_instance(iss, tag):
    proc = find_test_instance(iss, tag)
    if proc:
        proc.kill()


def kill_process(pid):
    proc = psutil.Process(pid)
    proc.kill()


def isrunning(iss, tag):
    proc = find_test_instance(iss, tag)
    if proc:
        return proc.pid
    else:
        return 0


def pid_isrunning(pid):
    proc = psutil.Process(pid)
    return proc


if __name__ == "__main__":
    # iss = sys.argv[1]
    # tag = sys.argv[2]

    inst = find_test_instances('op_test_tool.py')
    print(inst)



import json
import time
from future.backports.urllib.parse import parse_qs

from Cryptodome.PublicKey import RSA
from jwkest.ecc import P256
from jwkest.jwk import RSAKey
from jwkest.jwk import ECKey
from jwkest.jwk import SYMKey
from oic.extension import provider
from oic.extension.client import RegistrationRequest
from oic.oauth2.message import AccessTokenRequest
from oic.oauth2.message import ASConfigurationResponse
from oic.utils.keyio import keyjar_init

from otest.events import EV_PROTOCOL_REQUEST
from otest.events import EV_HTTP_RESPONSE

__author__ = 'roland'


class TestError(Exception):
    pass


def sort_string(string):
    if string is None:
        return ""

    _l = list(string)
    _l.sort()
    return "".join(_l)


class Provider(provider.Provider):
    def __init__(self, name, sdb, cdb, authn_broker, userinfo, authz,
                 client_authn, symkey, urlmap=None, ca_certs="", keyjar=None,
                 hostname="", template_lookup=None, template=None,
                 verify_ssl=True, capabilities=None, event_db=None,
                 config=None, **kwargs):

        provider.Provider.__init__(
            self, name, sdb, cdb, authn_broker, authz=authz,
            client_authn=client_authn, symkey=symkey, urlmap=urlmap,
            ca_bundle=ca_certs, keyjar=keyjar, hostname=hostname,
            verify_ssl=verify_ssl, capabilities=capabilities, config=config,
            **kwargs)

        self.claims_type = ["normal"]
        self.behavior_type = []
        self.server.behavior_type = self.behavior_type
        self.claim_access_token = {}
        self.init_keys = []
        self.update_key_use = ""
        self.events = event_db
        self.userinfo = userinfo
        self.template_lookup = template_lookup
        self.template = template
        self.strict = False

    def create_providerinfo(self, pcr_class=ASConfigurationResponse,
                            setup=None):
        _response = provider.Provider.create_providerinfo(self, pcr_class,
                                                          setup)

        if "isso" in self.behavior_type:
            _response["issuer"] = "https://example.com"

        return _response

    def check_scheme(self, reg_req):
        # Do initial verification that all endpoints from the client uses https
        for endp in ["redirect_uris", "jwks_uri", "initiate_login_uri"]:
            try:
                uris = reg_req[endp]
            except KeyError:
                continue

            if not isinstance(uris, list):
                uris = [uris]
            for uri in uris:
                if not uri.startswith("https://"):
                    return self._error(
                        error="invalid_configuration_parameter",
                        descr="Non-HTTPS endpoint in '{}'".format(endp))


    def registration_endpoint(self, request, authn=None, **kwargs):
        try:
            reg_req = RegistrationRequest().deserialize(request, "json")
        except ValueError:
            reg_req = RegistrationRequest().deserialize(request)

        self.events.store(EV_PROTOCOL_REQUEST, reg_req)

        if self.strict:
            resp = self.check_scheme(reg_req)
            if resp:
                return resp

        _response = provider.Provider.registration_endpoint(
            self, request=request, authn=authn, **kwargs)

        self.init_keys = []
        if "jwks_uri" in reg_req:
            if _response.status == "200 OK":
                # find the client id
                req_resp = ASConfigurationResponse().from_json(
                    _response.message)
                for kb in self.keyjar[req_resp["client_id"]]:
                    if kb.imp_jwks:
                        self.trace.info("Client JWKS: {}".format(kb.imp_jwks))

        return _response

    def authorization_endpoint(self, request="", cookie=None, **kwargs):
        _req = parse_qs(request)

        _response = provider.Provider.authorization_endpoint(self, request,
                                                             cookie, **kwargs)

        # This is just for logging purposes
        try:
            _resp = self.server.http_request(_req["request_uri"][0])
        except KeyError:
            pass
        else:
            if _resp.status_code == 200:
                self.events.store(EV_HTTP_RESPONSE, _resp)

        return _response

    def token_endpoint(self, request="", authn=None, **kwargs):

        try:
            req = AccessTokenRequest().deserialize(request, 'urlencoded')
            client_id = self.client_authn(self, req, authn)
        except Exception as err:
            self.trace.error("Failed to verify client due to: %s" % err)
            return self._error(error="incorrect_behavior",
                               descr="Failed to verify client")

        if self.events:
            self.events.store(EV_PROTOCOL_REQUEST, req)

        try:
            self._update_client_keys(client_id)
        except TestError:
            return self._error(error="incorrect_behavior",
                               descr="No change in client keys")

        _response = provider.Provider.token_endpoint(
            self, authn=authn, request=request, **kwargs)

        return _response

    def generate_jwks(self, mode):
        if "rotenc" in self.behavior_type:  # Rollover encryption keys
            rsa_key = RSAKey(kid="rotated_rsa_{}".format(time.time()),
                             use="enc").load_key(RSA.generate(2048))
            ec_key = ECKey(kid="rotated_ec_{}".format(time.time()),
                           use="enc").load_key(P256)

            keys = [rsa_key.serialize(private=True),
                    ec_key.serialize(private=True)]
            new_keys = {"keys": keys}
            #self.do_key_rollover(new_keys, "%d")

            signing_keys = [k.to_dict() for k in self.keyjar.get_signing_key()]
            new_keys["keys"].extend(signing_keys)
            return json.dumps(new_keys)
        elif "nokid1jwk" in self.behavior_type:
            alg = mode["sign_alg"]
            if not alg:
                alg = "RS256"
            keys = [k.to_dict() for kb in self.keyjar[""] for k in
                    list(kb.keys())]
            for key in keys:
                if key["use"] == "sig" and key["kty"].startswith(alg[:2]):
                    key.pop("kid", None)
                    jwk = dict(keys=[key])
                    return json.dumps(jwk)
            raise Exception(
                "Did not find sig {} key for nokid1jwk test ".format(alg))
        else:  # Return all keys
            keys = [k.to_dict() for kb in self.keyjar[""] for k in
                    list(kb.keys())]
            jwks = dict(keys=keys)
            return json.dumps(jwks)

    def _update_client_keys(self, client_id):
        if "updkeys" in self.behavior_type:
            if not self.init_keys:
                if "rp_enc_key" in self.baseurl:
                    self.update_key_use = "enc"
                else:
                    self.update_key_use = "sig"
                self.init_keys = []
                for kb in self.keyjar[client_id]:
                    for key in kb.available_keys():
                        if isinstance(key, SYMKey):
                            pass
                        elif key.use == self.update_key_use:
                            self.init_keys.append(key)
            else:
                for kb in self.keyjar[client_id]:
                    self.events.store("Updating client keys", '')
                    kb.update()
                    if kb.imp_jwks:
                        self.events.store("Client JWKS´", kb.imp_jwks)
                same = 0
                # Verify that the new keys are not the same
                for kb in self.keyjar[client_id]:
                    for key in kb.available_keys():
                        if isinstance(key, SYMKey):
                            pass
                        elif key.use == self.update_key_use:
                            if key in self.init_keys:
                                same += 1
                            else:
                                self.events.store("New key", key)
                if same == len(self.init_keys):  # no change
                    self.events.store("No change in keys")
                    raise TestError("Keys unchanged")
                else:
                    self.events.store('Key modifications',
                                      "{} changed, {} same".format(
                                          len(self.init_keys) - same, same))

    def __setattr__(self, key, value):
        if key == "keys":
            # Update the keyjar instead of just storing the keys description
            keyjar_init(self, value)
        else:
            super(provider.Provider, self).__setattr__(key, value)

    def _get_keyjar(self):
        return self.server.keyjar

    def _set_keyjar(self, item):
        self.server.keyjar = item

    keyjar = property(_get_keyjar, _set_keyjar)



import json
import os
import tarfile

from otest.check import CRITICAL
from otest.check import ERROR
from otest.check import INCOMPLETE
from otest.check import OK
from otest.check import STATUSCODE
from otest.check import WARNING
from otest.events import EV_CONDITION
from otest.events import EV_FAULT
from otest.events import layout

__author__ = 'roland'


def assert_summation(events, sid):
    status = OK
    result = []

    if events.get_data(EV_FAULT):
        status = ERROR
    else:
        for test_result in events.get_data(EV_CONDITION):
            result.append('{}'.format(test_result))
            if test_result.status > status:
                status = test_result.status

    info = {
        "id": sid,
        "status": status,
        "assertions": result
    }

    return info


def completed(events):
    """
    Figure out if the test ran to completion
    :param events: An otest.events.Events instance
    :return: True/False
    """
    for item in events.get_data(EV_CONDITION):
        if item.test_id == "Done" and item.status in [OK, ERROR]:
            return True

    return False


def eval_state(events):
    """
    The state of the test is equal to the worst status encountered
    :param events: An otest.events.Events instance
    :return: An integer representing a status code
    """
    if completed(events):
        res = OK
    else:
        res = INCOMPLETE

    if events.get_data(EV_FAULT):
        return ERROR  # Can't get worse

    for state in events.get_data(EV_CONDITION):
        if state.status > res:
            res = state.status

    return res


def get_errors(events):
    res = []
    for item in events.get_data(EV_FAULT):
        res.append(item.message)

    for item in events.get_data(EV_CONDITION):
        if item.status in [ERROR, CRITICAL]:
            res.append(item.message)

    return '. '.join(res)


def result_code(events):
    _state = eval_state(events)
    if _state == INCOMPLETE:
        tag = "PARTIAL RESULT"
    else:
        if _state < WARNING:
            tag = "PASSED"
        else:
            tag = STATUSCODE[_state]

    return tag


def represent_result(events):
    """
    A textual representation of the status of the test result
    :param events: An otest.events.Events instance
    :return: A text string
    """

    tag = result_code(events)

    lines = [tag]

    errors = []
    warning = []
    for state in events.get_data(EV_CONDITION):
        if state.status == ERROR:
            if state.message:
                errors.append('{}'.format(state.message))
        elif state.status == WARNING:
            if state.message:
                warning.append('{}'.format(state.message))

    if errors:
        lines.append('Errors:')
        lines.append("\n".join(errors))

    if warning:
        lines.append('Warnings:')
        lines.append("\n".join(warning))

    text = "\n".join(lines)

    return text


# -----------------------------------------------------------------------------

def trace_output(events):
    """

    """
    start = 0
    element = ["Trace output\n"]
    for event in events:
        if not start:
            start = event.timestamp
        element.append(layout(start, event))
    element.append("\n")
    return element


def condition(events, html=False):
    """

    """
    if html:
        element = [""]
    else:
        element = ["Conditions\n"]
    for cond in events.get_data(EV_CONDITION):
        element.append('{}'.format(cond))
    if html:
        return "\n".join(element)
    else:
        element.append("\n")
        return element


def pprint_json(json_txt):
    _jso = json.loads(json_txt)
    return json.dumps(_jso, sort_keys=True, indent=2, separators=(',', ': '))


def mk_tar_dir(issuer, test_profile):
    wd = os.getcwd()

    # Make sure there is a tar directory
    tardirname = wd
    for part in ["tar", issuer, test_profile]:
        tardirname = os.path.join(tardirname, part)
        if not os.path.isdir(tardirname):
            os.mkdir(tardirname)

    # Now walk through the log directory and make symlinks from
    # the log files to links in the tar directory
    logdirname = os.path.join(wd, "log", issuer, test_profile)
    for item in os.listdir(logdirname):
        if item.startswith("."):
            continue

        ln = os.path.join(logdirname, item)
        tn = os.path.join(tardirname, "{}.txt".format(item))

        if os.path.isfile(tn):
            os.unlink(tn)

        if not os.path.islink(tn):
            os.symlink(ln, tn)


def create_tar_archive(issuer, test_profile):
    mk_tar_dir(issuer, test_profile)

    wd = os.getcwd()
    _dir = os.path.join(wd, "tar", issuer)
    os.chdir(_dir)

    tar = tarfile.open("{}.tar".format(test_profile), "w")

    for item in os.listdir(test_profile):
        if item.startswith("."):
            continue

        fn = os.path.join(test_profile, item)

        if os.path.isfile(fn):
            tar.add(fn)
    tar.close()
    os.chdir(wd)



#!/usr/bin/env python
from oic.utils.keyio import create_and_store_rsa_key_pair

for name in ['pyoidc', '2nd', '3rd']:
    create_and_store_rsa_key_pair("keys/{}_enc".format(name), size=2048)
    create_and_store_rsa_key_pair("keys/{}_sig".format(name), size=2048)



import json
from otest import Unknown

__author__ = 'roland'


def translate(spec, func_factories):
    asp = {}
    for key, args in spec.items():
        _fnc = None
        for fact in func_factories:
            try:
                _fnc = fact(key)
            except Exception:
                pass
            else:
                break
        if not _fnc:
            raise Unknown('Check {}'.format(key))
        asp[_fnc] = args
    return asp


COPY = ['descr']


def parse_json_conf(cnf_file, cls_factories, chk_factories, func_factories):
    """

    :param cnf_file:
    :param cls_factories:
    :param chk_factory:
    :return:
    """
    stream = open(cnf_file, 'r')
    js = json.load(stream)
    stream.close()
    res = {}
    for tid, spec in js.items():
        ops = {}
        for oper, asse in spec.items():
            if oper == 'setup':
                if asse:
                    ops['setup'] = translate(asse, func_factories)
                else:
                    ops['setup'] = {}
            elif oper in COPY:
                ops[oper] = asse
            else:
                _cls = ''
                for fact in cls_factories:
                    try:
                        _cls = fact(oper)
                    except Exception:
                        pass
                    else:
                        break
                if not _cls:
                    raise Unknown('Operation {}'.format(oper))

                ops[_cls.__name__] = {'assert': {}}
                if 'assert' in asse and asse['assert'] is not None:
                    ops[_cls.__name__]['assert'] = translate(
                        asse['assert'], chk_factories)
        res[tid] = ops
    return res



import copy
import logging

from otest import Done
from otest.parse_cnf import sort

__author__ = 'roland'

logger = logging.getLogger(__name__)


def pmap(sprof, prof):
    try:
        p = sprof.split('.')
    except AttributeError:
        p = sprof

    try:
        rt = prof.split('.')
    except AttributeError:
        rt = prof

    if rt[0] == '*' or p[0] in rt:
        return True
    else:
        return False


class Node(object):
    def __init__(self, name, desc, mti=None):
        self.name = name
        self.desc = desc
        self.mti = mti
        self.state = 0
        self.info = ""
        self.rmc = False
        self.experr = False
        self.complete = False


class SessionHandler(object):
    def __init__(self, flows=None, order=None, tool_version='', **kwargs):
        self.test_flows = flows
        self.order = order
        self.extra = kwargs
        self.tool_version = tool_version
        self._dict = {}

    @property
    def profile(self):
        return self.extra['profile']

    def session_setup(self, path="", flow=None, index=0):
        logger.info("session_setup")

        _keys = list(self.keys())
        for key in _keys:
            if key in ["tests", "flow_names", "response_type",
                       "test_info", "profile", 'test_conf', 'sid']:
                continue
            else:
                del self[key]

        self["testid"] = path
        if not flow:
            flow = self.test_flows.expanded_conf(path)

        self['flow'] = flow
        self["sequence"] = self["flow"]["sequence"]
        self["sequence"].append(Done)
        self["index"] = index

    def init_session(self, profile=None):
        if profile is None:
            profile = self.profile

        self["tests"] = self.test_flows.matches_profile(profile)
        return self._dict

    def reset_session(self, profile=None):
        _keys = list(self.keys())
        for key in _keys:
            if key.startswith("_"):
                continue
            else:
                del self[key]
        self.init_session(profile)

    def session_init(self):
        if "tests" not in self or self['tests'] == []:
            self.init_session()
            return True
        else:
            return False

    def dump(self, filename):
        pass

    def load(self, filename):
        pass

    def keys(self):
        return self._dict.keys()

    def update(self, new):
        self._dict.update(new)

    def __delitem__(self, item):
        del self._dict[item]

    def __getitem__(self, item):
        return self._dict[item]

    def __setitem__(self, key, value):
        self._dict[key] = value

    def __contains__(self, item):
        return item in self._dict

    def items(self):
        return self._dict.items()



from oic.utils.time_util import in_a_while

__author__ = 'roland'


RT = {"C": "code", "D": "client cred", "T": "token"}
ATTR = ["profile"]


def to_profile(session, representation="list"):
    p = session["profile"].split(".")
    prof = [RT[p[0]]]

    if representation == "list":
        return prof
    elif representation == "dict":
        ret = {}
        for r in range(0, len(prof)):
            ret[ATTR[r]] = prof[r]
        return ret


def get_profile_info(session, test_id=None):
    try:
        _conv = session["conv"]
    except KeyError:
        pass
    else:
        try:
            iss = _conv.entity.provider_info["issuer"]
        except (TypeError, KeyError):
            iss = ""

        profile = to_profile(session, "dict")

        if test_id is None:
            try:
                test_id = session["testid"]
            except KeyError:
                return {}

        return {"Issuer": iss, "Profile": profile, "Test ID": test_id,
                "Test description": session.test_flows[test_id]['desc'],
                "Timestamp": in_a_while()}

    return {}


def make_client(**kw_args):
    """
    Have to get own copy of keyjar

    :param kw_args:
    :return:
    """
    c_keyjar = kw_args["keyjar"].copy()
    _cli = Client(client_authn_method=CLIENT_AUTHN_METHOD, keyjar=c_keyjar)

    c_info = {'keyjar': c_keyjar}
    for arg, val in list(kw_args.items()):
        if arg in ['keyjar']:
            continue
        setattr(_cli, arg, val)
        c_info[arg] = val

    return _cli, c_info

# def get_check(check_id):
#     package = ocheck
#     prefix = package.__name__ + "."
#     for importer, modname, ispkg in pkgutil.iter_modules(package.__path__,
#                                                          prefix):
#         module = __import__(modname, fromlist="dummy")
#         chk = module.factory(check_id)
#         if chk:
#             return chk
#
#     return aa_check.factory(check_id)



from otest import prof_util

__author__ = 'roland'

RT = {"C": "code", "T": "token", 'D': 'client_credentials'}


class ProfileHandler(prof_util.ProfileHandler):
    def to_profile(self, representation="list"):
        p = self.session.profile.split('.')
        prof = RT[p[0]]

        if representation == "list":
            return [prof]
        elif representation == "dict":
            return {'response_type': prof}



import json
import logging
import csv
import pkg_resources

from oic.extension.token import JWTToken

from oic.utils.authn.authn_context import AuthnBroker
from oic.utils.authn.client import verify_client
from oic.utils.authz import AuthzHandling
from oic.utils.keyio import keyjar_init
from oic.utils.sdb import SessionDB
from oic.utils.userinfo import UserInfo

from otest.events import Events
from otest.rp.provider import Provider

logger = logging.getLogger(__name__)

__author__ = 'roland'


def read_uri_schemes(filename):
    csvfile = open(filename, 'r')
    l = csvfile.readline()
    l = l.strip()
    fieldnames = l.split(',')
    reader = csv.DictReader(csvfile, fieldnames)
    return dict(
        [(r['URI Scheme'], '{} {}'.format(r['Description'], r['Reference'])) for
         r in reader])


def read_path2port_map(filename):
    """
    Reads csv file containing two columns: column1 is path name,
     column2 is port number

    :param filename:
    :return: dictionary with port as key and path as value
    """
    res = {}
    with open(filename, 'r') as csvfile:
        reader = csv.reader(csvfile)
        for row in reader:
            res[row[1]] = row[0]
    return res


def as_arg_setup(args, lookup, config):
    if args.port:
        _port = args.port
    else:
        if args.tls:
            _port = 443
        else:
            _port = 80

    if args.path2port:
        # means there is a reverse proxy in front translating
        # path -> port
        p2p_map = read_path2port_map(args.path2port)
        _path = p2p_map[_port]
        if args.xport:
            _issuer = "{base}:{port}/{path}".format(base=config.baseurl,
                                                    port=args.xport,
                                                    path=_path)
            _port = args.xport
        else:
            _issuer = "{base}/{path}".format(base=config.baseurl, path=_path)
    else:  # the old port based
        _path = ''
        _issuer = "{base}:{port}".format(base=config.baseurl, port=_port)
        if args.tls and _issuer.startswith('http://'):
            _issuer = _issuer.replace('http://', 'https://')

    cdb = {}

    ac = AuthnBroker()

    for authkey, value in list(config.AUTHENTICATION.items()):
        authn = None
        # if "UserPassword" == authkey:
        #     from oic.utils.authn.user import UsernamePasswordMako
        #     authn = UsernamePasswordMako(None, "login.mako", LOOKUP, PASSWD,
        #                                  "authorization")

        if "NoAuthn" == authkey:
            from oic.utils.authn.user import NoAuthn

            authn = NoAuthn(None, user=config.AUTHENTICATION[authkey]["user"])

        if authn is not None:
            ac.add(config.AUTHENTICATION[authkey]["ACR"], authn,
                   config.AUTHENTICATION[authkey]["WEIGHT"])

    # dealing with authorization
    authz = AuthzHandling()

    if config.USERINFO == "SIMPLE":
        # User info is a simple dictionary in this case statically defined in
        # the configuration file
        userinfo = UserInfo(config.USERDB)
    else:
        userinfo = None

    as_args = {
        "name": _issuer,
        'instance_path': _path,
        'instance_port': _port,
        "cdb": cdb,
        "authn_broker": ac,
        "userinfo": userinfo,
        "authz": authz,
        "client_authn": verify_client,
        "symkey": config.SYM_KEY,
        "template_lookup": lookup,
        "template": {"form_post": "form_response.mako"},
        "jwks_name": "./static/jwks_{}.json",
        'event_db': Events(),
    }

    try:
        as_args['behavior'] = config.BEHAVIOR
    except AttributeError:
        pass

    com_args = {
        "baseurl": config.baseurl,
    }

    for arg in ['name', 'cdb', 'authn_broker', 'userinfo', 'authz', 'template',
                'jwks_name', 'client_authn', 'symkey', 'template_lookup']:
        com_args[arg] = as_args[arg]

    # Add own keys for signing/encrypting JWTs
    try:
        # a throw-away OP used to do the initial key setup
        _op = Provider(sdb=SessionDB(com_args["baseurl"]), **com_args)
        jwks = keyjar_init(_op, config.keys)
    except KeyError:
        key_arg = {}
    else:
        key_arg = {"jwks": jwks, "keys": config.keys}
        as_args['jwks_name'] = 'static/jwks.json'
        f = open('static/jwks.json', 'w')
        f.write(json.dumps(jwks))
        f.close()

        if args.insecure:
            _op.keyjar.verify_ssl = False
        else:
            _op.keyjar.verify_ssl = True

        as_args['keyjar'] = _op.keyjar
        as_args['sdb'] = SessionDB(
            com_args["baseurl"],
            token_factory=JWTToken('T', keyjar=_op.keyjar,
                                   lt_pattern={'code': 3600, 'token': 900},
                                   iss=com_args['baseurl'],
                                   sign_alg='RS256'),
            refresh_token_factory=JWTToken(
                'R', keyjar=_op.keyjar, lt_pattern={'': 24 * 3600},
                iss=com_args['baseurl'])
        )

    return as_args, key_arg


def main_setup(args, lookup, config):
    config.issuer = config.issuer % args.port
    config.SERVICE_URL = config.SERVICE_URL % args.port

    as_args, key_arg = as_arg_setup(args, lookup, config)

    kwargs = {
        "template_lookup": lookup,
        "template": {"form_post": "form_response.mako"},
    }

    # Should I care about verifying the certificates used by other entities
    if args.insecure:
        kwargs["verify_ssl"] = False
    else:
        kwargs["verify_ssl"] = True

    op_arg = key_arg

    try:
        op_arg["cookie_ttl"] = config.COOKIETTL
    except AttributeError:
        pass

    try:
        op_arg["cookie_name"] = config.COOKIENAME
    except AttributeError:
        pass

    # print URLS
    if args.debug:
        op_arg["debug"] = True

    # # All endpoints the OpenID Connect Provider should answer on
    # add_endpoints(ENDPOINTS)
    # op_arg["endpoints"] = ENDPOINTS

    if args.port == 80:
        _baseurl = config.baseurl
    else:
        if config.baseurl.endswith("/"):
            config.baseurl = config.baseurl[:-1]
        _baseurl = "%s:%d" % (config.baseurl, args.port)

    if not _baseurl.endswith("/"):
        _baseurl += "/"

    op_arg["baseurl"] = _baseurl

    logger.info('setup kwargs: {}'.format(kwargs))

    try:
        op_arg["marg"] = multi_keys(as_args, config.multi_keys)
    except AttributeError as err:
        pass

    op_arg['uri_schemes'] = read_uri_schemes(
        pkg_resources.resource_filename('otest', 'uri-schemes-1.csv'))

    if args.op_profiles:
        profiles = {}
        for p in args.op_profiles:
            profiles.update(json.loads(open(p).read()))
    else:
        profiles = {}

    op_arg['profiles'] = profiles

    logger.info("setup as_args: {}".format(as_args))
    logger.info(" --   op_arg: {}".format(op_arg))

    return as_args, op_arg, config


def multi_keys(as_args, key_conf):
    # a throw-away OP used to do the initial key setup
    _op = Provider(**as_args)
    jwks = keyjar_init(_op, key_conf, "m%d")

    return {"jwks": jwks, "keys": key_conf}



import os

from future.backports.urllib.parse import quote_plus

from otest.utils import with_or_without_slash
from otest.time_util import in_a_while

RESPONSE = 0
WEBFINGER = 1
DISCOVER = 2
REGISTER = 3
CRYPTO = 4
EXTRAS = 5
FORMPOST = 6

LABEL = ['return_type', 'webfinger', 'discover', 'register', 'crypto',
         'extra', "form_post"]

__author__ = 'roland'

EMAP = {'s': 'sig', 'n': 'none', 'e': 'enc'}
EKEYS = list(EMAP.keys())
EKEYS.sort()  # Make the result deterministic

PKEYS = LABEL[:]
PKEYS.remove('crypto')
PKEYS.extend(list(EMAP.values()))

RT = {"C": "code", "I": "id_token", "T": "token", "CT": "code token",
      'CI': 'code id_token', 'IT': 'id_token token',
      'CIT': 'code id_token token'}
RT_INV = dict([(v, k) for k, v in RT.items()])

WF = {'T': 'webfinger', 'F': 'no-webfinger'}
OC = {"T": "discovery", "F": "no-discovery"}
REG = {"T": "dynamic", "F": "static"}
CR = {"n": "none", "s": "sign", "e": "encrypt"}
EX = {"+": "extras"}
FP = {'T': 'form_post'}

ATTR = ["return_type", "webfinger", "openid-configuration", "registration",
        "crypto", "extras", "form_post"]


def simplify_return_type(spec):
    if ',' in spec:
        p = spec.split(',')
    elif ' ' in spec:
        p = spec.split(' ')
    else:
        p = [spec]
    p.sort()
    return RT_INV[' '.join(p)]


def abbr_return_type(spec):
    return RT[spec]


def verify_profile(profile):
    p = profile.split('.')
    if len(p) < 4:
        return False
    if p[0] not in ['C', 'I', 'IT', 'CT', 'CIT', 'CI']:
        return False
    for i in range(1, 4):
        if p[i] not in ['F', 'T']:
            return False
    return True


def from_profile(code):
    # Of the form <typ>.<webf>.<disc>.<reg>.*['+'/'n'/'s'/'se']
    # for example:
    # C.T.T.T..  - code response_type, webfinger & dynamic discovery &
    #                                   registration
    # CIT.F.T.F.. - response_type=["code","id_token","token"],
    #               No webfinger support,
    #               does dynamic discovery
    #               and static client registration

    p = code.split('.')

    _prof = {"return_type": p[RESPONSE],
             "extra": False,
             "sig": True,
             'enc': False,
             'none': False,
             'form_post': False}

    for x, y in {WEBFINGER: 'webfinger', DISCOVER: 'discover',
                 REGISTER: 'register'}.items():
        try:
            _prof[y] = (p[x] == 'T')
        except (KeyError, IndexError):
            _prof[y] = True

    if len(p) > CRYPTO:
        for k, v in EMAP.items():
            if k in p[CRYPTO]:
                _prof[v] = True

        if len(p) > EXTRAS:
            if '+' in p[EXTRAS]:
                _prof['extra'] = True

            if len(p) > FORMPOST:
                if 'T' in p[FORMPOST]:
                    _prof['form_post'] = True

    return _prof


def to_profile(pdict):
    code = [pdict["return_type"]]

    for key in ["webfinger", "discover", "register"]:
        try:
            if pdict[key]:
                code.append("T")
            else:
                code.append("F")
        except KeyError:
            code.append("F")

    # Do the rest backwards
    tail = []
    for typ, val in [('form_post', 'T'), ('extra', '+')]:
        try:
            _fp = pdict[typ]
        except KeyError:
            tail.append('')
        else:
            if _fp:
                tail.append(val)
            else:
                tail.append('')

    ext = ''
    for k in EKEYS:
        try:
            if pdict[EMAP[k]]:
                ext += k
        except KeyError:
            pass

    if ext:
        tail.append(ext)
    else:
        tail.append('')

    i = 0
    while tail[i] == '':
        i += 1
        if i == len(tail):
            break

    if i == len(tail):
        pass
    else:
        tail = tail[i:]
        tail.reverse()
        code.extend(tail)

    return ".".join(code)


def repr_profile(profile, representation="list", with_webfinger=True):
    """

    :param profile: Expected to be list of items
    :param representation: Which type of output that is expected
    :param with_webfinger: Is WebFinger specification included
    :return:
    """
    prof = ["+".join([RT[x] for x in profile[0]])]
    if with_webfinger:
        _spec = [WF, OC, REG]
    else:
        _spec = [OC, REG]

    i = 0
    for tag_text in _spec:
        i += 1
        try:
            prof.append("%s" % tag_text[profile[i]])
        except IndexError:
            pass

    try:
        i += 1
        _val = "%s" % "+".join([CR[x] for x in profile[i]])
    except (KeyError, IndexError):
        pass
    else:
        if _val:
            prof.append(_val)
        i += 1
        try:
            _val = profile[i]
        except (KeyError, IndexError):
            pass
        else:
            if _val:
                prof.append("%s" % EX[_val])
            i += 1
            try:
                _val = profile[i]
            except (KeyError, IndexError):
                pass
            else:
                if _val:
                    prof.append("%s" % FP[_val])

    if representation == "list":
        return prof
    elif representation == "dict":
        ret = {}
        for r in range(0, len(prof)):
            ret[ATTR[r]] = prof[r]

        if "extras" in ret:
            ret["extras"] = True
        return ret


def do_registration(profile):
    if profile.split('.')[REGISTER] == 'F':
        return False
    else:
        return True


def do_discovery(profile):
    if profile.split('.')[DISCOVER] == 'F':
        return False
    else:
        return True


def return_type(profile):
    return profile.split('.')[RESPONSE]


def update_profile(old, new):
    oa = from_profile(old)
    na = from_profile(new)
    oa.update(na)
    return to_profile(oa)


def compress_profile(item):
    _prof = to_profile(item)
    item['profile'] = _prof
    for attr in PKEYS:
        try:
            del item[attr]
        except KeyError:
            pass
    return item


def expand_profile(item):
    try:
        _prof = from_profile(item['profile'])
    except KeyError:
        pass
    else:
        del item['profile']
        item.update(_prof)

    return item


def _cmp_prof(a, b):
    """

    :param a: list of strings
    :param b: list of strings
    :return: True/False if a maps to b
    """
    # basic, implicit, hybrid
    if b[RESPONSE] != "":
        if a[RESPONSE] not in b[RESPONSE].split(','):
            return False

    try:
        # dynamic discovery & registry
        for n in [WEBFINGER, DISCOVER, REGISTER]:
            if b[n] != "":
                if a[n] != b[n]:
                    return False
    except IndexError:
        print("Too short a:{}, b:{}".format(a, b))
        raise

    if len(a) > CRYPTO:
        if len(b) > CRYPTO:
            if b[CRYPTO] != '':
                if not set(a[CRYPTO]).issuperset(set(b[CRYPTO])):
                    return False

    for var in [EXTRAS, FORMPOST]:
        if len(b) > var:
            if len(a) > var:
                if a[var] != b[var]:
                    return False
            else:
                return False

    return True


def map_prof(a, b):
    """
    Checks that the demands in b are met by a

    :param a:
    :param b:
    :return: True/False
    """
    if a == b:
        return True

    if isinstance(b, list):
        return _cmp_prof(a, b)
    elif '.' in b:
        b = b.split('.')
        if '.' in a:
            a = a.split('.')
        return _cmp_prof(a, b)
    else:
        if b == '*':
            return True
        else:
            bl = b.split(',')
            if isinstance(a, list):
                if a[0] in bl:
                    return True
                else:
                    return False
            elif a in bl:
                return True
            else:
                return False


class ProfileHandler(object):
    def __init__(self, session):
        self.session = session

    @staticmethod
    def webfinger(profile):
        return profile[WEBFINGER] == "T"

    @staticmethod
    def discover(profile):
        return profile[DISCOVER] == "T"

    @staticmethod
    def register(profile):
        return profile[REGISTER] == "T"

    @staticmethod
    def form_post(profile):
        return profile[FORMPOST] == "T"

    def to_profile(self, representation="list"):
        return []

    def get_profile_info(self, test_id=None):
        try:
            _conv = self.session["conv"]
        except KeyError:
            pass
        else:
            try:
                iss = _conv.entity.provider_info["issuer"]
            except (TypeError, KeyError):
                iss = ""

            profile = self.to_profile("dict")

            if test_id is None:
                try:
                    test_id = self.session["testid"]
                except KeyError:
                    return {}

            return {
                "Issuer": iss, "Profile": profile,
                "Test ID": test_id,
                "Test description": self.session.test_flows[test_id]["desc"],
                "Timestamp": in_a_while()}

        return {}

    def log_path(self, **kwargs):
        _conv = self.session["conv"]

        try:
            iss = _conv.entity.provider_info["issuer"]
        except (TypeError, KeyError):
            return ""
        else:
            qiss = quote_plus(iss)

        path = with_or_without_slash(os.path.join("log", qiss))
        if path is None:
            path = os.path.join("log", qiss)

        prof = ".".join(self.to_profile())

        if not os.path.isdir("{}/{}".format(path, prof)):
            os.makedirs("{}/{}".format(path, prof))

        if 'test_id' not in kwargs:
            _test_id = self.session["testid"]
        else:
            _test_id = kwargs['test_id']

        return "{}/{}/{}".format(path, prof, _test_id)


class SimpleProfileHandler(ProfileHandler):
    @staticmethod
    def webfinger(profile):
        return True

    @staticmethod
    def discover(profile):
        return True

    @staticmethod
    def register(profile):
        return True

    @staticmethod
    def form_post(profile):
        return True

    def get_profile_info(self, test_id=None):
        try:
            _conv = self.session["conv"]
        except KeyError:
            pass
        else:
            try:
                iss = _conv.entity.provider_info["issuer"]
            except AttributeError:
                iss = _conv.entity.baseurl
            except (TypeError, KeyError):
                iss = ""

            profile = RT[''.join(self.session["profile"])]

            if test_id is None:
                try:
                    test_id = self.session["testid"]
                except KeyError:
                    return {}

            return {
                "Issuer": iss, "Profile": profile,
                "Test ID": test_id,
                "Test description": self.session.test_flows[test_id]["desc"],
                "Timestamp": in_a_while()}
        return {}

    def log_path(self, **kwargs):
        path = os.path.join("log", kwargs['sid'])

        prof = ".".join(self.to_profile())

        if not os.path.isdir("{}/{}".format(path, prof)):
            os.makedirs("{}/{}".format(path, prof))

        try:
            _test_id = kwargs['test_id']
        except KeyError:
            _test_id = self.session["testid"]

        return "{}/{}/{}".format(path, prof, _test_id)



import logging
import os

# from urllib.parse import unquote
from future.backports.urllib.parse import unquote

from oic.utils.http_util import NotFound
from oic.utils.http_util import Response

from otest.check import ERROR
from otest.check import OK
from otest.check import WARNING
from otest.check import INCOMPLETE
from otest.handling import InfoHandling
from otest.utils import with_or_without_slash
from otest.summation import represent_result

__author__ = 'roland'

logger = logging.getLogger(__name__)

TEST_RESULTS = {OK: "OK", ERROR: "ERROR", WARNING: "WARNING",
                INCOMPLETE: "INCOMPLETE"}


class WebIh(InfoHandling):
    def __init__(self, conf=None, flows=None, profile_handler=None,
                 lookup=None, cache=None, environ=None, start_response=None,
                 session=None, base_url='', **kwargs):
        InfoHandling.__init__(self, flows=flows,
                              profile_handler=profile_handler,
                              cache=cache, session=session, **kwargs)

        self.conf = conf
        self.lookup = lookup
        self.environ = environ
        self.start_response = start_response
        self.base_url = base_url

    def flow_list(self):
        try:
            resp = Response(mako_template="flowlist.mako",
                            template_lookup=self.lookup,
                            headers=[])
        except Exception as err:
            logger.error(err)
            raise

        argv = {
            "flows": self.flows.display_info(self.session['tests']),
            "profile": self.session["profile"],
            # "test_info": list(self.session["test_info"].keys()),
            "base": self.base_url,
        }

        return resp(self.environ, self.start_response, **argv)

    def profile_edit(self):
        resp = Response(mako_template="profile.mako",
                        template_lookup=self.lookup,
                        headers=[])

        argv = {'profile': self.session["profile"],
                "base": self.base_url}

        return resp(self.environ, self.start_response, **argv)

    def test_info(self, testid):
        resp = Response(mako_template="testinfo.mako",
                        template_lookup=self.lookup,
                        headers=[])

        info = self.flows.test_info[testid]

        argv = {
            "profile": info["profile_info"],
            "events": info['events'],
            "result": represent_result(info['events']).replace("\n", "<br>\n"),
            'base': self.base_url
        }

        return resp(self.environ, self.start_response, **argv)

    def not_found(self):
        """Called if no URL matches."""
        resp = NotFound()
        return resp(self.environ, self.start_response)

    def static(self, path):
        logger.info("[static]sending: %s" % (path,))

        try:
            text = open(path, 'rb').read()
            logger.debug('Read {} bytes'.format(len(text)))
            if path.endswith(".ico"):
                self.start_response('200 OK', [('Content-Type',
                                                "image/x-icon")])
            elif path.endswith(".html"):
                self.start_response('200 OK', [('Content-Type', 'text/html')])
            elif path.endswith(".json"):
                self.start_response('200 OK', [('Content-Type',
                                                'application/json')])
            elif path.endswith(".jwt"):
                self.start_response('200 OK', [('Content-Type',
                                                'application/jwt')])
            elif path.endswith(".txt"):
                self.start_response('200 OK', [('Content-Type', 'text/plain')])
            elif path.endswith(".css"):
                self.start_response('200 OK', [('Content-Type', 'text/css')])
            else:
                self.start_response('200 OK', [('Content-Type', "text/plain")])
            return [text]
        except IOError:
            resp = NotFound()
            return resp(self.environ, self.start_response)

    def _display(self, root, issuer, profile):
        item = []
        logger.debug('curdir:{}'.format(os.getcwd()))
        if profile:
            path = os.path.join(root, issuer, profile).replace(":", "%3A")
            argv = {"issuer": unquote(issuer), "profile": profile}

            path = with_or_without_slash(path)
            if path is None:
                resp = Response("No saved logs")
                return resp(self.environ, self.start_response)

            for _name in os.listdir(path):
                if _name.startswith("."):
                    continue
                fn = os.path.join(path, _name)
                if os.path.isfile(fn):
                    item.append((unquote(_name), fn))
        else:
            if issuer:
                argv = {'issuer': unquote(issuer), 'profile': ''}
                path = os.path.join(root, issuer).replace(":", "%3A")
            else:
                argv = {'issuer': '', 'profile': ''}
                path = root

            path = with_or_without_slash(path)
            if path is None:
                resp = Response("No saved logs")
                return resp(self.environ, self.start_response)

            for _name in os.listdir(path):
                if _name.startswith("."):
                    continue
                fn = os.path.join(path, _name)
                if os.path.isdir(fn):
                    item.append((unquote(_name), os.path.join(path, _name)))

        resp = Response(mako_template="logs.mako",
                        template_lookup=self.lookup,
                        headers=[])

        item.sort()
        argv["logs"] = item
        argv["base"] = self.base_url
        return resp(self.environ, self.start_response, **argv)

    def display_log(self, root, issuer="", profile="", testid=""):
        """

        :param root:
        :param issuer:
        :param profile:
        :param testid:
        :return:
        """
        logger.info(
            "display_log root: '{root}' issuer: '{iss}', profile: '{prof}' "
            "testid: '{tid}'".format(
                root=root, iss=issuer, prof=profile, tid=testid))
        if testid:
            path = os.path.join(root, issuer, profile, testid).replace(":",
                                                                       "%3A")
            return self.static(path)
        else:
            if issuer:
                return self._display(root, issuer, profile)
            else:
                resp = Response("No saved logs")
                return resp(self.environ, self.start_response)

    def err_response(self, where, err):
        self._store_error(where, err)
        return self.flow_list()

    def sorry_response(self, homepage, err):
        resp = Response(mako_template="sorry.mako",
                        template_lookup=self.lookup,
                        headers=[])
        argv = {
            "htmlpage": homepage,
            "error": str(err),
            'base': self.base_url
        }
        return resp(self.environ, self.start_response, **argv)

    def opresult(self, conv):
        return self.flow_list()

    def opresult_fragment(self):
        resp = Response(mako_template="opresult_repost.mako",
                        template_lookup=self.lookup,
                        headers=[])
        argv = {'base': self.base_url}
        return resp(self.environ, self.start_response, **argv)

    def respond(self, resp):
        if isinstance(resp, Response):
            return resp(self.environ, self.start_response)
        else:
            return resp



import importlib
import json
import logging
import sys

from mako.lookup import TemplateLookup
from oic.utils.keyio import build_keyjar

from otest.parse_cnf import parse_yaml_conf
from otest.rp.setup import read_path2port_map
from otest.utils import SERVER_LOG_FOLDER
from otest.utils import setup_logging

logger = logging.getLogger(__name__)


def construct_app_args(args, oper, func, default_profiles):
    """

    :param args: Command arguments, argparse instance
    :param oper: Operations module
    :param func: Functions module
    :param default_profiles: The default profiles module
    :return: Application arguments
    """
    sys.path.insert(0, ".")
    CONF = importlib.import_module(args.config)

    #setup_logging("%s/rp_%s.log" % (SERVER_LOG_FOLDER, CONF.PORT), logger)

    fdef = {'Flows': {}, 'Order': [], 'Desc': {}}
    cls_factories = {'': oper.factory}
    func_factory = func.factory

    for flow_def in args.flows:
        spec = parse_yaml_conf(flow_def, cls_factories, func_factory)
        fdef['Flows'].update(spec['Flows'])
        fdef['Desc'].update(spec['Desc'])
        fdef['Order'].extend(spec['Order'])

    if args.profiles:
        profiles = importlib.import_module(args.profiles)
    else:
        profiles = default_profiles

    try:
        if args.operations:
            operations = importlib.import_module(args.operations)
        else:
            operations = oper
    except AttributeError:
        operations = oper

    # Add own keys for signing/encrypting JWTs
    jwks, keyjar, kidd = build_keyjar(CONF.KEYS)

    try:
        if args.staticdir:
            _sdir = args.staticdir
        else:
            _sdir = 'static'
    except AttributeError:
        _sdir = 'static'

    # If this instance is behind a reverse proxy or on its own
    if CONF.BASE.endswith('/'):
        CONF.BASE = CONF.BASE[:-1]
    if args.path2port:
        ppmap = read_path2port_map(args.path2port)
        _path = ppmap[str(CONF.PORT)]
        if args.xport:
            _port = CONF.PORT
            _base = '{}:{}/{}/'.format(CONF.BASE, str(CONF.PORT), _path)
        else:
            _base = '{}/{}/'.format(CONF.BASE, _path)
            if args.tls:
                _port = 443
            else:
                _port = 80
    else:
        _port = CONF.PORT
        if _port not in [443, 80]:
            _base = '{}:{}'.format(CONF.BASE, _port)
        else:
            _base = CONF.BASE
        _path = ''

    # -------- JWKS ---------------

    if args.path2port:
        jwks_uri = "{}{}/jwks_{}.json".format(_base, _sdir, _port)
        f = open('{}/jwks_{}.json'.format(_sdir, _port), "w")
    elif _port not in [443, 80]:
        jwks_uri = "{}:{}/{}/jwks_{}.json".format(CONF.BASE, _port, _sdir,
                                                  _port)
        f = open('{}/jwks_{}.json'.format(_sdir, _port), "w")
    else:
        jwks_uri = "{}/{}/jwks.json".format(CONF.BASE, _sdir)
        f = open('{}/jwks.json'.format(_sdir), "w")
    f.write(json.dumps(jwks))
    f.close()

    # -------- MAKO setup -----------
    try:
        if args.makodir:
            _dir = args.makodir
            if not _dir.endswith("/"):
                _dir += "/"
        else:
            _dir = "./"
    except AttributeError:
        _dir = './'

    LOOKUP = TemplateLookup(directories=[_dir + 'templates', _dir + 'htdocs'],
                            module_directory=_dir + 'modules',
                            input_encoding='utf-8',
                            output_encoding='utf-8')

    _client_info = CONF.CLIENT

    # Now when the basci URL for the RP is constructed update the
    # redirect_uris and the post_logout_redirect_uris
    try:
        ri = _client_info['registration_info']
    except KeyError:
        pass
    else:
        ri['redirect_uris'] = [r.format(_base) for r in ri['redirect_uris']]
        try:
            ri['post_logout_redirect_uris'] = [r.format(_base) for r in
                                               ri['post_logout_redirect_uris']]
        except KeyError:
            pass

    _base += '/'
    _client_info.update(
        {"base_url": _base, 'client_id': _base, "kid": kidd, "keyjar": keyjar,
         "jwks_uri": jwks_uri}
    )

    if args.insecure:
        _client_info['verify_ssl'] = False

    # Test profile either as a command line argument or if not that
    # from the configuration file
    if args.profile:
        _profile = args.profile
    else:
        _profile = CONF.TOOL['profile']

    # Application arguments
    app_args = {
        "flows": fdef['Flows'], "conf": CONF,
        "client_info": _client_info, "order": fdef['Order'],
        "profiles": profiles, "operation": operations, "cache": {},
        "profile": _profile, "lookup": LOOKUP, "desc": fdef['Desc'],
    }

    return _path, app_args



from otest.events import EV_SEND, EV_RUN
from otest.events import EV_REDIRECT_URL
from otest.events import EV_HTTP_RESPONSE
from otest.operation import Operation
from otest.verify import Verify
from oic.oauth2.message import Message
from oic.utils import http_util

__author__ = 'roland'


class Response(Operation):
    endpoint = ''
    message_cls = Message

    def __init__(self, conv, inut, sh, **kwargs):
        Operation.__init__(self, conv, inut, sh, **kwargs)
        self.expect_error = {}
        self.req_args = {}
        self.op_args = {}
        self.csi = None
        self.entity = self.conv.entity
        self.relay_state = ''
        self.request_id = ''
        self.response_args = {}
        self.request_inst = None

    def construct_message(self):
        return {}

    def handle_request(self, *args):
        raise NotImplemented

    def __call__(self, *args, **kwargs):
        if self.skip:
            return
        else:
            cls_name = self.__class__.__name__
            if self.tests["pre"] or self.tests["post"]:
                _ver = Verify(self.check_factory, self.conv, cls_name=cls_name)
                _ver.test_sequence(self.tests["pre"])

            self.conv.events.store(EV_RUN,cls_name)

            res = self.run(*args, **kwargs)

            if res:
                return res

    def run(self, *args, **kwargs):
        #self.handle_request()
        send_args = self.construct_message()
        if isinstance(send_args, http_util.Response):
            return send_args

        self.conv.events.store(EV_SEND, send_args)
        self.conv.events.store(EV_REDIRECT_URL, send_args['url'])
        res = self.entity.send(**send_args)
        self.conv.events.store(EV_HTTP_RESPONSE, res)
        return res


class RedirectResponse(Response):
    _class = None
    _args = {}
    _method = 'GET'

    def run(self):
        self.handle_request()
        send_args = self.construct_message()
        if isinstance(send_args, Response):
            return send_args

        self.conv.events.store(EV_SEND, send_args)
        self.conv.events.store(EV_REDIRECT_URL, send_args['url'])
        res = self.entity.send(**send_args)
        self.conv.events.store(EV_HTTP_RESPONSE, res)
        return res



import logging
import os

SERVER_LOG_FOLDER = "server_log"


def setup_logging(logfile, log):
    hdlr = logging.FileHandler(logfile)
    base_formatter = logging.Formatter(
        "%(asctime)s %(name)s:%(levelname)s %(message)s")

    hdlr.setFormatter(base_formatter)
    log.addHandler(hdlr)
    log.setLevel(logging.DEBUG)


def not_logging(logfile, logger):
    hdlr = logging.FileHandler(logfile)
    base_formatter = logging.Formatter(
        "%(asctime)s %(name)s:%(levelname)s %(message)s")

    hdlr.setFormatter(base_formatter)
    logger.addHandler(hdlr)
    logger.setLevel(logging.DEBUG)


def get_test_info(session, test_id):
    return session["test_info"][test_id]


def setup_common_log():
    common_logger = logging.getLogger("common")
    hdlr = logging.FileHandler("%s/common.log" % SERVER_LOG_FOLDER)
    base_formatter = logging.Formatter(
        "%(asctime)s %(name)s:%(levelname)s %(message)s")
    hdlr.setFormatter(base_formatter)
    common_logger.addHandler(hdlr)
    common_logger.setLevel(logging.DEBUG)

    return common_logger, hdlr, base_formatter


def with_or_without_slash(path):
    if os.path.isdir(path):
        return path

    if path.endswith("%2F"):
        path = path[:-3]
        if os.path.isdir(path):
            return path
    else:
        path += "%2F"
        if os.path.isdir(path):
            return path

    return None



import json


class JLog(object):
    def __init__(self, logger, sid, tag='JLOG'):
        self.logger = logger
        self.id = sid
        self.tag = tag

    def _msg(self, info):
        return '<<{}:{}>> {}'.format(self.tag, self.id, json.dumps(info))

    def info(self, info):
        self.logger.info(self._msg(info))

    def debug(self, info):
        self.logger.debug(self._msg(info))

    def exception(self, info):
        self.logger.exception(self._msg(info))

    def error(self, info):
        self.logger.error(self._msg(info))

    def warning(self, info):
        self.logger.warning(self._msg(info))


preline = "<h3>A list of {} that are saved on disk for this {}:</h3>"


def display_log(logs, issuer, profile, base):
    el = []
    if issuer:
        if profile:
            el.append(preline.format('tests', 'profile'))
        else:
            el.append(preline.format('profiles', 'issuer'))
    else:
        el.append(preline.format('issuers', 'test server'))

    el.append("<ul>")

    if profile:
        for name, path in logs:
            el.append(
                '<li><a href="{}{}" download="{}.html">{}</a>'.format(
                    base, path, name, name))
    elif 'issuer':
        for name, path in logs:
            _tarfile = "{}{}.tar".format(base, path.replace("log", "tar"))
            el.append(
                '<li><a href="{}{}">{}</a> tar file:<a href="{}">'.format(
                    base, path, name, _tarfile))
            el.append('Download logs</a>')
    else:
        for name, path in logs:
            el.append('<li><a href="{}{}">{}</a>'.format(base, path, name))
    el.append("</ul>")
    return '\n'.join(el)



#!/usr/bin/python
#
# Copyright (C) 2016 Catalogix, Sweden
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import re

from setuptools import setup, find_packages

__author__ = 'roland'

with open('src/otest/__init__.py', 'r') as fd:
    version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]',
                        fd.read(), re.MULTILINE).group(1)


setup(
    version = version,
    name="otest",
    description="Basic Test framework for testing OAuth2/OIDC Authorization "
                "server and relaying party implementations",
    author="Roland Hedberg",
    author_email="roland@catalogix.se",
    license="Apache 2.0",
    packages=find_packages('src'),
    package_dir={"": "src"},
    data_files=[('otest', ['data/uri-schemes-1.csv'])],
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3.5",
        "Topic :: Software Development :: Libraries :: Python Modules"],
    install_requires=[
        "argparse",
        "requests >= 2.0.0",
        'pyYAML',
        'pycryptodomex',
        'oic >= 0.12.0',
        'pyjwkest >= 1.3.2',
        'future',
        'entropy',
        'cherrypy',
        'robobrowser',
        'psutil'
    ],
    zip_safe=False,
    scripts=['script/testtool.py', 'script/make_keys.py']
)



from datetime import datetime
from datetime import timedelta

__author__ = 'roland'

TIME_FORMAT = "%Y-%m-%dT%H:%M:%SZ"


def time_in_a_while(days=0, seconds=0, microseconds=0, milliseconds=0,
                    minutes=0, hours=0, weeks=0):
    """
    format of timedelta:
        timedelta([days[, seconds[, microseconds[, milliseconds[,
                    minutes[, hours[, weeks]]]]]]])
    :return: UTC time
    """
    delta = timedelta(days, seconds, microseconds, milliseconds,
                      minutes, hours, weeks)
    return datetime.utcnow() + delta


def in_a_while(days=0, seconds=0, microseconds=0, milliseconds=0,
               minutes=0, hours=0, weeks=0, time_format=TIME_FORMAT):
    """
    format of timedelta:
        timedelta([days[, seconds[, microseconds[, milliseconds[,
                    minutes[, hours[, weeks]]]]]]])
    """
    if time_format is None:
        time_format = TIME_FORMAT

    return time_in_a_while(days, seconds, microseconds, milliseconds,
                           minutes, hours, weeks).strftime(time_format)





import functools
import inspect
import json
import logging
import os
import sys
import time

from Cryptodome.PublicKey import RSA
from future.backports.urllib.parse import urlparse
from jwkest.jwk import RSAKey
from oic.oauth2.message import Message
from oic.utils.keyio import KeyBundle
from oic.utils.keyio import dump_jwks
from oic.utils.keyio import ec_init
from otest import RequirementsNotMet
from otest import Unknown
from otest import operation
from otest.aus.request import SyncGetRequest
from otest.operation import request_with_client_http_session

__author__ = 'roland'

logger = logging.getLogger(__name__)


def include(url, test_id):
    p = urlparse(url)
    if p.path[1:].startswith(test_id):
        if len(p.path[1:].split("/")) <= 1:
            return os.path.join(url, "_/_/_/normal")
        else:
            return url

    return "%s://%s/%s%s_/_/_/normal" % (p.scheme, p.netloc, test_id, p.path)


def get_id_token(responses):
    """
    Find the id_tokens issued, last one first in the list
    :param responses: A list of Response instance, text message tuples
    :return: list of IdTokens instances
    """
    res = []
    for resp, txt in responses:
        try:
            res.insert(0, resp["id_token"])
        except KeyError:
            pass
    return res


class Operation(operation.Operation):
    message_cls = Message

    def __init__(self, conv, inut, sh, test_id='', conf=None,
                 funcs=None, check_factory=None, cache=None, profile='',
                 tool_conf=None):
        operation.Operation.__init__(self, conv, inut, sh, test_id,
                                     conf, funcs, check_factory, cache,
                                     tool_conf)

        try:
            self.profile = profile.split('.')
        except AttributeError:
            self.profile = profile

        # Monkey-patch: make sure we use the same http session (preserving
        # cookies) when fetching keys from issuers 'jwks_uri' as for the
        # rest of the test sequence
        import oic.utils.keyio

        oic.utils.keyio.request = functools.partial(
            request_with_client_http_session, self)


class UpdateProviderKeys(Operation):
    def __call__(self, *args, **kwargs):
        issuer = self.conv.entity.provider_info["issuer"]
        # Update all keys
        for keybundle in self.conv.entity.keyjar.issuer_keys[issuer]:
            keybundle.update()


class RotateKey(Operation):
    def __call__(self):
        keyjar = self.conv.entity.keyjar
        self.conv.entity.original_keyjar = keyjar.copy()

        # invalidate the old key
        old_kid = self.op_args["old_kid"]
        old_key = keyjar.get_key_by_kid(old_kid)
        old_key.inactive_since = time.time()

        # setup new key
        key_spec = self.op_args["new_key"]
        typ = key_spec["type"].upper()
        if typ == "RSA":
            kb = KeyBundle(keytype=typ, keyusage=key_spec["use"])
            kb.append(RSAKey(use=key_spec["use"]).load_key(
                RSA.generate(key_spec["bits"])))
        elif typ == "EC":
            kb = ec_init(key_spec)
        else:
            raise Exception('Wrong key type')

        # add new key to keyjar with
        list(kb.keys())[0].kid = self.op_args["new_kid"]
        keyjar.add_kb("", kb)

        # make jwks and update file
        keys = []
        for kb in keyjar[""]:
            keys.extend(
                [k.to_dict() for k in list(kb.keys()) if not k.inactive_since])
        jwks = dict(keys=keys)
        with open(self.op_args["jwks_path"], "w") as f:
            f.write(json.dumps(jwks))


class RestoreKeyJar(Operation):
    def __call__(self):
        self.conv.entity.keyjar = self.conv.entity.original_keyjar

        # make jwks and update file
        keys = []
        for kb in self.conv.entity.keyjar[""]:
            keys.extend([k.to_dict() for k in list(kb.keys())])
        jwks = dict(keys=keys)
        with open(self.op_args["jwks_path"], "w") as f:
            f.write(json.dumps(jwks))


class ReadRegistration(SyncGetRequest):
    def op_setup(self):
        _client = self.conv.entity
        self.req_args["access_token"] = _client.registration_access_token
        self.op_args["authn_method"] = "bearer_header"
        self.op_args["endpoint"] = _client.registration_response[
            "registration_client_uri"]


class FetchKeys(Operation):
    def __call__(self):
        kb = KeyBundle(source=self.conv.entity.provider_info["jwks_uri"])
        kb.verify_ssl = False
        kb.update()

        try:
            self.conv.keybundle.append(kb)
        except AttributeError:
            self.conv.keybundle = [kb]


class RotateKeys(Operation):
    def __init__(self, conv, inut, sh, **kwargs):
        Operation.__init__(self, conv, inut, sh, **kwargs)
        self.jwk_name = "export/jwk.json"
        self.new_key = {}
        self.kid_template = "_%d"
        self.key_usage = ""

    def __call__(self):
        # find the name of the file to which the JWKS should be written
        try:
            _uri = self.conv.entity.registration_response["jwks_uri"]
        except KeyError:
            raise RequirementsNotMet("No dynamic key handling")

        r = urlparse(_uri)
        # find the old key for this key usage and mark that as inactive
        for kb in self.conv.entity.keyjar.issuer_keys[""]:
            for key in list(kb.keys()):
                if key.use in self.new_key["use"]:
                    key.inactive = True

        kid = 0
        # only one key
        _nk = self.new_key
        _typ = _nk["type"].upper()

        if _typ == "RSA":
            kb = KeyBundle(source="file://%s" % _nk["key"],
                           fileformat="der", keytype=_typ,
                           keyusage=_nk["use"])
        else:
            kb = {}

        for k in list(kb.keys()):
            k.serialize()
            k.kid = self.kid_template % kid
            kid += 1
            self.conv.entity.kid[k.use][k.kty] = k.kid
        self.conv.entity.keyjar.add_kb("", kb)

        dump_jwks(self.conv.entity.keyjar[""], r.path[1:])


class RotateSigKeys(RotateKeys):
    def __init__(self, conv, inut, sh, **kwargs):
        RotateKeys.__init__(self, conv, inut, sh, **kwargs)
        self.new_key = {"type": "RSA", "key": "./keys/second_sig.key",
                        "use": ["sig"]}
        self.kid_template = "sig%d"


class RotateEncKeys(RotateKeys):
    def __init__(self, conv, inut, sh, **kwargs):
        RotateKeys.__init__(self, conv, inut, sh, **kwargs)
        self.new_key = {"type": "RSA", "key": "./keys/second_enc.key",
                        "use": ["enc"]}
        self.kid_template = "enc%d"


class Cache(Operation):
    pass


def factory(name):
    for fname, obj in inspect.getmembers(sys.modules[__name__]):
        if inspect.isclass(obj):
            if name == fname:
                return obj

    from otest import operation

    obj = operation.factory(name)
    if not obj:
        raise Unknown("Couldn't find the operation: '{}'".format(name))
    return obj



import json
import re

# from urllib.parse import urlparse
from future.backports.urllib.parse import urlparse
from otest.events import EV_FUNCTION
from robobrowser import RoboBrowser

NO_CTRL = "No submit control with the name='%s' and value='%s' could be found"


class FlowException(Exception):
    def __init__(self, function="", content="", url=""):
        Exception.__init__(self)
        self.function = function
        self.content = content
        self.url = url

    def __str__(self):
        return json.dumps(self.__dict__)


class InteractionNeeded(Exception):
    pass


def no_func():
    return None


class RResponse():
    """
    A Response class that behaves in the way that mechanize expects it.
    Links to a requests.Response
    """

    def __init__(self, resp):
        self._resp = resp
        self.index = 0
        self.text = resp.text
        if isinstance(self.text, str):
            if resp.encoding.upper() == "UTF-8":
                self.text = self.text.encode("utf-8")
            else:
                self.text = self.text.encode("latin-1")
        self._len = len(self.text)
        self.url = str(resp.url)
        self.statuscode = resp.status_code

    def geturl(self):
        return self._resp.url

    def __getitem__(self, item):
        try:
            return getattr(self._resp, item)
        except AttributeError:
            return getattr(self._resp.headers, item)

    def __getattribute__(self, item):
        try:
            return getattr(self._resp, item)
        except AttributeError:
            return getattr(self._resp.headers, item)

    def read(self, size=0):
        """
        Read from the content of the response. The class remembers what has
        been read so it's possible to read small consecutive parts of the
        content.

        :param size: The number of bytes to read
        :return: Somewhere between zero and 'size' number of bytes depending
            on how much it left in the content buffer to read.
        """
        if size:
            if self._len < size:
                return self.text
            else:
                if self._len == self.index:
                    part = None
                elif self._len - self.index < size:
                    part = self.text[self.index:]
                    self.index = self._len
                else:
                    part = self.text[self.index:self.index + size]
                    self.index += size
                return part
        else:
            return self.text


class Interaction(object):
    def __init__(self, httpc, interactions=None, verify_ssl=True):
        self.httpc = httpc
        self.browser = RoboBrowser()
        self.interactions = interactions
        self.verify_ssl = verify_ssl

    def pick_interaction(self, response, base):
        if self.interactions is None:
            return None

        self.browser._update_state(response)
        _bs = self.browser.parsed
        unic = ""

        for interaction in self.interactions:
            _match = 0
            for attr, val in list(interaction["matches"].items()):
                if attr == "url":
                    if val == base:
                        _match += 1
                elif attr == "title":
                    if _bs is None:
                        break
                    if _bs.title is None:
                        break
                    if val in _bs.title.contents:
                        _match += 1
                    else:
                        _c = _bs.title.contents
                        if isinstance(_c, list) and not isinstance(_c, str):
                            for _line in _c:
                                if val in _line:
                                    _match += 1
                                    continue
                elif attr == "content":
                    if unic and val in unic:
                        _match += 1

            if _match == len(interaction["matches"]):
                return interaction

        raise InteractionNeeded("No interaction matched")

    def pick_form(self, forms, **kwargs):
        """
        Picks which form in a web-page that should be used

        :param forms: A list of robobrowser.Forms instances
        :return: The picked form or None if no form matched the criteria.
        """

        _form = None

        if len(forms) == 1:
            _form = forms[0]
        else:
            if "pick" in kwargs:
                _dict = kwargs["pick"]
                for form in forms:
                    if _form:
                        break
                    for key, _ava in list(_dict.items()):
                        if key == "form":
                            _keys = list(form.attrs.keys())
                            for attr, val in list(_ava.items()):
                                if attr in _keys and val == form.attrs[attr]:
                                    _form = form
                        elif key == "control":
                            prop = _ava["id"]
                            _default = _ava["value"]
                            try:
                                orig_val = form[prop]
                                if isinstance(orig_val, str):
                                    if orig_val == _default:
                                        _form = form
                                elif _default in orig_val:
                                    _form = form
                            except KeyError:
                                pass
                            except Exception as err:
                                pass
                        elif key == "method":
                            if form.method == _ava:
                                _form = form
                        else:
                            _form = None

                        if not _form:
                            break
            elif "index" in kwargs:
                _form = forms[int(kwargs["index"])]

        return _form

    def select_form(self, response, **kwargs):
        """
        Pick a form on a web page, possibly enter some information and submit
        the form.

        :param orig_response: The original response (as returned by requests)
        :return: The response do_click() returns
        """
        self.browser._update_state(response)
        forms = self.browser.get_forms()
        form = self.pick_form(forms, **kwargs)

        if not forms:
            raise Exception("Can't pick a form !!")

        if "set" in kwargs:
            for key, val in list(kwargs["set"].items()):
                if key.startswith("_"):
                    continue
                if "click" in kwargs and kwargs["click"] == key:
                    continue

                try:
                    form[key].value = val
                except (ValueError):
                    pass
                except Exception as err:
                    raise
                    # cntrl = form.find_control(key)
                    # if isinstance(cntrl, ListControl):
                    #     form[key] = [val]
                    # else:
                    #     raise

        if form.action in kwargs["tester"].my_endpoints():
            _res = {}
            for name, cnt in form.fields.items():
                _res[name] = cnt.value
            return _res

        try:
            requests_args = kwargs["requests_args"]
        except KeyError:
            requests_args = {}

        self.browser.submit_form(form, **requests_args)
        return self.browser.state.response

    # noinspection PyUnusedLocal
    def chose(self, orig_response, path, **kwargs):
        """
        Sends a HTTP GET to a url given by the present url and the given
        relative path.

        :param orig_response: The original response
        :param content: The content of the response
        :param path: The relative path to add to the base URL
        :return: The response do_click() returns
        """

        if not path.startswith("http"):
            try:
                _url = orig_response.url
            except KeyError:
                _url = kwargs["location"]

            part = urlparse(_url)
            url = "%s://%s%s" % (part[0], part[1], path)
        else:
            url = path

        return self.httpc.send(url, "GET")
        # return resp, ""

    def redirect(self, orig_response, url_regex, **kwargs):
        """
        Simulates a JavaScript redirect by extracting the target of the
        redirection from the page content using the given regex

        :param orig_response: The original response
        :param url_regex: The regex that defines how the target of the redirect
                          can be extracted from the content
        """

        matches = re.findall(url_regex, orig_response.content)
        no_of_matches = len(matches)
        if not no_of_matches == 1:
            raise InteractionNeeded("Expected single match but found %d",
                                    no_of_matches)

        url = matches[0]
        return self.httpc.send(url, "GET")

    def post_form(self, response, **kwargs):
        """
        The same as select_form but with no possibility of changing the content
        of the form.

        :param response: The original response (as returned by requests)
        :return: The response submit_form() returns
        """

        form = self.pick_form(response, **kwargs)

        return self.browser.submit_form(form)

    def response(self, response, **kwargs):
        return {"text": response.text}

    # noinspection PyUnusedLocal
    def interaction(self, args):
        _type = args["type"]
        if _type == "form":
            return self.select_form
        elif _type == "link":
            return self.chose
        elif _type == "response":
            return self.response
        elif _type == "redirect":
            return self.redirect
        elif _type == "javascript_redirect":
            return self.redirect
        else:
            return no_func


# ========================================================================


class Action(object):
    def __init__(self, args):
        self.args = args or {}
        self.request = None

    def update(self, dic):
        self.args.update(dic)

    # noinspection PyUnusedLocal
    def post_op(self, result, conv, args):
        pass

    def __call__(self, tester, location, response, features, **kwargs):
        _conv = tester.conv
        intact = _conv.interaction
        function = intact.interaction(self.args)

        try:
            _args = self.args.copy()
        except (KeyError, AttributeError):
            _args = {}

        _args["_events_"] = _conv.events
        _args["location"] = location
        _args["features"] = features
        _args["tester"] = tester
        _args["requests_args"] = kwargs

        if _conv.events:
            _conv.events.store(EV_FUNCTION,
                               {'name': function.__name__, 'kwargs': _args})

        result = function(response, **_args)
        self.post_op(result, _conv, _args)
        return result



import logging
from otest.events import EV_OPERATION
from otest.summation import assert_summation
from otest.verify import Verify


def setup_logger(log, log_file_name="tt.log"):
    # logger = logging.getLogger("")
    hdlr = logging.FileHandler(log_file_name)
    base_formatter = logging.Formatter(
        "%(asctime)s %(name)s:%(levelname)s %(message)s")

    hdlr.setFormatter(base_formatter)
    log.addHandler(hdlr)
    log.setLevel(logging.DEBUG)


def make_list(flows, profile, map_prof, **kw_args):
    f_names = list(flows.keys())
    f_names.sort()
    flow_names = []
    for k in kw_args["order"]:
        k += '-'
        l = [z for z in f_names if z.startswith(k)]
        flow_names.extend(l)

    res = []
    if profile:
        sprofile = profile.split(".")
    else:
        sprofile = []

    for tid in flow_names:
        _flow = flows[tid]

        if map_prof(sprofile, _flow["profile"].split(".")):
            res.append(tid)

    return res


def make_entity(**kw_args):
    raise NotImplemented()


def run_flow(profiles, conv, test_id, conf, profile, chk_factory, index=0):
    print(("==" + test_id))
    conv.test_id = test_id
    conv.conf = conf

    if index >= len(conv.flow["sequence"]):
        return None

    conv.index = index

    for item in conv.flow["sequence"][index:]:
        if isinstance(item, tuple):
            cls, funcs = item
        else:
            cls = item
            funcs = {}

        _oper = cls(conv=conv, profile=profile, test_id=test_id,
                    conf=conf, funcs=funcs, chk_factory=chk_factory)
        conv.operation = _oper
        conv.events.store(EV_OPERATION, cls.__name__, sender='run_flow')
        _oper.setup(profiles.PROFILEMAP)
        _oper()

        conv.index += 1

    try:
        if conv.flow["tests"]:
            _ver = Verify(chk_factory, conv)
            _ver.test_sequence(conv.flow["tests"])
    except KeyError:
        pass
    except Exception as err:
        raise

    info = assert_summation(conv, test_id)

    return info



from otest.aus.preproc import PMAP
from otest.prof_util import CRYPTO
from otest.prof_util import EMAP
from otest.prof_util import EXTRAS
from otest.prof_util import FORMPOST

PMAPL = list(PMAP.keys())
PMAPL.sort()

L2I = {"discovery": 1, "registration": 2}
LONG_NAMES = {"none": "none", "sig": "signing", "enc": "encryption"}

RADIO = '<div class="radio"><label><input type="radio" name="{}" value="{}">{}</label></div>'
RADIO_C = '<div class="radio"><label><input type="radio" name="{}" value="{}" checked>{}</label></div>'
CHECK = '<div class="checkbox"><label><input type="checkbox" name="{}">{}</label></div>'
CHECK_C = '<div class="checkbox"><label><input type="checkbox" name="{}" checked>{}</label></div>'


def profile_form(prof):
    p = prof.split(".")
    el = []
    for key in PMAPL:
        txt = PMAP[key]
        if key == p[0]:
            el.append(RADIO_C.format('return_type', key, txt))
        else:
            el.append(RADIO.format('return_type', key, txt))
    el.append("<br>")
    el.append("These you can't change here:")
    el.append("<ul>")
    for mode in ["discovery", "registration"]:
        if p[L2I[mode]] == "T":
            el.append("<li>Dynamic %s" % mode)
        else:
            el.append("<li>Static %s" % mode)

    el.append(
        "</ul><p>Response mode form_post support:</br>")
    el.append(
        "<i>(make sure to register https://&lt;host&gt;:&lt;port&gt;/authz_post as (one of) the redirect URI(s) when enabled!)</i></li></br>")
    if len(p) > FORMPOST and p[FORMPOST] == "T":
        el.append(CHECK_C.format('form_post', ''))
    else:
        el.append(CHECK.format('form_post', ''))
    el.append('</p>')

    el.append("</ul><p>Cryptographic support:<br>")
    if len(p) > CRYPTO:
        vs = p[CRYPTO]
    else:
        vs = ''

    for code, attr in EMAP.items():
        if code in vs:
            el.append(CHECK_C.format(attr, LONG_NAMES[attr]))
        else:
            el.append(CHECK.format(attr, LONG_NAMES[attr]))
    el.append("</p>")

    el.append(
        '</ul><p>Check this if you want extra tests (not needed for any '
        'certification profiles): ')
    if len(p) > EXTRAS and p[EXTRAS] == "+":
        el.append(CHECK_C.format('extra', ''))
    else:
        el.append(CHECK.format('extra', ''))
    el.append('</p>')
    return "\n".join(el)



import logging
import os

# from urllib.parse import unquote
import cherrypy
from future.backports.urllib.parse import unquote

from otest.aus.preproc.flow_list import display_profile
from otest.aus.preproc.flow_list import legends
from otest.aus.preproc.flow_list import op_choice
from otest.aus.preproc.logs import display_log
from otest.aus.preproc.profile import profile_form
from otest.aus.preproc.testinfo import do_assertions
from otest.aus.preproc.testinfo import profile_output
from otest.aus.preproc.testinfo import trace_output

from otest.check import ERROR
from otest.check import OK
from otest.check import WARNING
from otest.check import INCOMPLETE
from otest.handling import InfoHandling
from otest.result import safe_url
from otest.utils import with_or_without_slash
from otest.summation import represent_result

__author__ = 'roland'

logger = logging.getLogger(__name__)

TEST_RESULTS = {OK: "OK", ERROR: "ERROR", WARNING: "WARNING",
                INCOMPLETE: "INCOMPLETE"}


class WebIh(InfoHandling):
    def __init__(self, conf=None, flow_state=None, profile_handler=None,
                 cache=None, session=None, base_url='',
                 pre_html=None, **kwargs):
        InfoHandling.__init__(self, flow_state=flow_state,
                              profile_handler=profile_handler,
                              cache=cache, session=session, **kwargs)

        self.conf = conf
        self.base_url = base_url
        self.pre_html = pre_html

    def flow_list(self):
        try:
            _tests = self.session['tests']
        except KeyError:
            _tests = self.flow_state.matches_profile(self.session.profile)
            self.session['tests'] = _tests

        _msg = self.pre_html['flow_list.html'].format(
            display_profile=display_profile(self.session.profile),
            op_choice=op_choice(
                self.base_url,
                self.flow_state.display_info(_tests)),
            legends=legends(),
            version=self.session.tool_version
        )
        return _msg

    def profile_edit(self):
        _msg = self.pre_html["profile.html"].format(
            profile_form=profile_form(self.session.profile),
            version=self.session.tool_version,
            base=self.base_url)
        return _msg

    def test_info(self, testid):
        try:
            info = self.flow_state.test_info[testid]
        except KeyError:
            _sess = self.session

            _l = ["You have not run this test during this test session.<br>"]
            _su = safe_url(_sess.iss)
            path0 = "log/{}/{}/{}/{}".format(_su, _sess.tag, _sess.profile,
                                             testid)
            if _su.endswith('2F'):
                path1 = "log/{}/{}/{}/{}".format(_su[:-2], _sess.tag,
                                                 _sess.profile, testid)
            else:
                path1 = "log/{}2F/{}/{}/{}".format(_su, _sess.tag,
                                                   _sess.profile, testid)

            for _path in [path0, path1]:
                if os.path.isfile(_path):
                    _l.extend([
                        "If you want to see a previous run you can look here ",
                        '<a href="/{}">link</a><br>'.format(_path)])

            return self.pre_html['sorry.html'].format(
                error='\n'.join(_l),
                link='<a href="{}">link</a>'.format(self.base_url)
            )
        else:
            _events = info['events']
            _msg = self.pre_html['testinfo.html'].format(
                profile_output=profile_output(
                    info["profile_info"],
                    version=self.session.tool_version),
                trace_output=trace_output(_events),
                result=represent_result(_events).replace("\n", "<br>\n"),
                assertions=do_assertions(_events),
                base=self.base_url,
                version=self.session.tool_version
            )
            return _msg

    def _display(self, root, issuer, profile):
        item = []
        logger.debug('curdir:{}'.format(os.getcwd()))
        if profile:
            path = os.path.join(root, issuer, profile).replace(":", "%3A")
            argv = {"issuer": unquote(issuer), "profile": profile}

            path = with_or_without_slash(path)
            if path is None:
                return "No saved logs"

            for _name in os.listdir(path):
                if _name.startswith("."):
                    continue
                fn = os.path.join(path, _name)
                if os.path.isfile(fn):
                    item.append((unquote(_name), fn))
        else:
            if issuer:
                argv = {'issuer': unquote(issuer), 'profile': ''}
                path = os.path.join(root, issuer).replace(":", "%3A")
            else:
                argv = {'issuer': '', 'profile': ''}
                path = root

            path = with_or_without_slash(path)
            if path is None:
                return b'No saved logs'

            for _name in os.listdir(path):
                if _name.startswith("."):
                    continue
                fn = os.path.join(path, _name)
                if os.path.isdir(fn):
                    item.append((unquote(_name), os.path.join(path, _name)))

        item.sort()
        _msg = self.pre_html['logs.html'].format(
            display_log=display_log(logs=item, base=self.base_url, **argv),
            version=self.session.tool_version
        )
        return _msg

    def display_log(self, root, issuer="", profile="", testid=""):
        """

        :param root:
        :param issuer:
        :param profile:
        :param testid:
        :return:
        """
        logger.info(
            "display_log root: '{root}' issuer: '{iss}', profile: '{prof}' "
            "testid: '{tid}'".format(
                root=root, iss=issuer, prof=profile, tid=testid))

        if testid:
            path = os.path.join(root, issuer, profile, testid).replace(":",
                                                                       "%3A")
            raise cherrypy.HTTPRedirect(path)
        else:
            if issuer:
                return self._display(root, issuer, profile)
            else:
                return b'No saved logs'

    def err_response(self, where, err):
        self._store_error(where, err)
        return self.flow_list()

    @staticmethod
    def reset_link(url):
        return "<a href='%sreset'>link</a>" % url

    def sorry_response(self, homepage, err):
        _msg = self.pre_html['sorry.html'].format(
            error=str(err),
            link=self.reset_link(self.base_url)
        )
        return _msg

    def opresult_fragment(self):
        _msg = self.pre_html['opresult_repost.html']
        return _msg

    def respond(self, resp):
        return resp



# -*- coding: utf-8 -*-
"""
    otest
    ~~~~~~~~~~

    Basic functionality needed by OAuth2, OIDC and UMA test tools
    :copyright: (c) 2016 by Roland Hedberg.
    :license: APACHE 2.0, see LICENSE for more details.
"""
from __future__ import print_function

import json
import logging
import time
import traceback
import requests
import sys
from subprocess import Popen, PIPE
from oic.oauth2 import HttpError

from otest.events import EV_END

__author__ = 'roland'
__version__ = '0.7.5'
__license__ = 'Apache 2.0'
__copyright__ = 'Copyright 2019 Roland Hedberg'


LOCAL_PATH = "export/"
END_TAG = "==== END ===="
CRYPTSUPPORT = {"none": "n", "signing": "s", "encryption": "e"}


class AATestError(Exception):
    pass


class FatalError(AATestError):
    pass


class Break(AATestError):
    pass


class Unknown(AATestError):
    pass


class ConfigurationError(AATestError):
    pass


class NotSupported(AATestError):
    pass


class RequirementsNotMet(AATestError):
    pass


class CheckError(AATestError):
    pass


class OperationError(AATestError):
    pass


class ConditionError(FatalError):
    pass


def as_bytes(s):
    """
    Convert an unicode string to bytes.
    :param s: Unicode / bytes string
    :return: bytes string
    """
    try:
        s = s.encode("utf-8", 'replace')
    except (AttributeError, UnicodeDecodeError):
        pass
    return s


def as_unicode(b):
    """
    Convert a byte string to a unicode string
    :param b: byte string
    :return: unicode string
    """
    try:
        b = b.decode()
    except (AttributeError, UnicodeDecodeError):
        pass
    return b


# class Trace(object):
#     def __init__(self, absolut_start=False):
#         self.trace = []
#         self.start = time.time()
#         if absolut_start:
#             self.trace.append("Trace init: {}".format(
#                 time.strftime("%a, %d %b %Y %H:%M:%S +0000", time.gmtime())))
#
#     @staticmethod
#     def format(resp):
#         raise NotImplemented()
#
#     def request(self, msg):
#         delta = time.time() - self.start
#         self.trace.append("%f --> %s" % (delta, msg))
#
#     def reply(self, msg):
#         delta = time.time() - self.start
#         self.trace.append("%f <-- %s" % (delta, msg))
#
#     def info(self, msg):
#         delta = time.time() - self.start
#         self.trace.append("%f %s" % (delta, msg))
#
#     def error(self, msg):
#         delta = time.time() - self.start
#         self.trace.append("%f [ERROR] %s" % (delta, msg))
#
#     def warning(self, msg):
#         delta = time.time() - self.start
#         self.trace.append("%f [WARNING] %s" % (delta, msg))
#
#     def __str__(self):
#         return "\n".join([as_unicode(t) for t in self.trace])
#
#     def clear(self):
#         self.trace = []
#
#     def __getitem__(self, item):
#         return self.trace[item]
#
#     def __next__(self):
#         for line in self.trace:
#             yield line
#
#     def lastline(self):
#         try:
#             return self.trace[-1]
#         except IndexError:
#             return ""
#
#     @staticmethod
#     def format(resp):
#         _d = {"claims": resp.to_dict()}
#         if resp.jws_header:
#             _d["jws header parameters"] = resp.jws_header
#         if resp.jwe_header:
#             _d["jwe header parameters"] = resp.jwe_header
#         return _d
#
#     def response(self, resp):
#         delta = time.time() - self.start
#         try:
#             cl_name = resp.__class__.__name__
#         except AttributeError:
#             cl_name = ""
#
#         if cl_name == "IdToken":
#             txt = json.dumps({"id_token": self.format(resp)},
#                              sort_keys=True, indent=2, separators=(',', ': '))
#             self.trace.append("%f %s: %s" % (delta, cl_name, txt))
#         else:
#             try:
#                 dat = resp.to_dict()
#             except AttributeError:
#                 txt = resp
#                 self.trace.append("%f %s" % (delta, txt))
#             else:
#                 if cl_name == "OpenIDSchema":
#                     cl_name = "UserInfo"
#                     if resp.jws_header or resp.jwe_header:
#                         dat = self.format(resp)
#                 elif "id_token" in dat:
#                     dat["id_token"] = self.format(resp["id_token"])
#
#                 txt = json.dumps(dat, sort_keys=True, indent=2,
#                                  separators=(',', ': '))
#
#                 self.trace.append("%f %s: %s" % (delta, cl_name, txt))


def start_script(path, wdir="", *args):
    if not path.startswith("/"):
        popen_args = ["./" + path]
    else:
        popen_args = [path]

    popen_args.extend(args)
    if wdir:
        return Popen(popen_args, stdout=PIPE, stderr=PIPE, cwd=wdir)
    else:
        return Popen(popen_args, stdout=PIPE, stderr=PIPE)


def stop_script_by_name(name):
    import subprocess
    import signal
    import os

    p = subprocess.Popen(['ps', '-A'], stdout=subprocess.PIPE)
    out, err = p.communicate()

    for line in out.splitlines():
        if name in line:
            pid = int(line.split(None, 1)[0])
            os.kill(pid, signal.SIGKILL)


def stop_script_by_pid(pid):
    import signal
    import os

    os.kill(pid, signal.SIGKILL)


def get_page(url):
    resp = requests.get(url)
    if resp.status_code == 200:
        return resp.text
    else:
        raise HttpError(resp.status_code)


def exception_trace(tag, exc, log=None):
    message = traceback.format_exception(*sys.exc_info())
    if log:
        if isinstance(exc, Exception):
            log.error("[%s] ExcList: %s" % (tag, "".join(message),))
        log.error("[%s] Exception: %s" % (tag, exc))
    else:
        if isinstance(exc, Exception):
            print("[%s] ExcList: %s" % (tag, "".join(message),),
                  file=sys.stderr)
        try:
            print("[%s] Exception: %s" % (tag, exc), file=sys.stderr)
        except UnicodeEncodeError:
            print("[%s] Exception: %s" % (
                tag, exc.message.encode("utf-8", "replace")), file=sys.stderr)
    return message


class ContextFilter(logging.Filter):
    """
    This is a filter which injects time laps information into the log.
    """

    def start(self):
        self.start = time.time()

    def filter(self, record):
        record.delta = time.time() - self.start
        return True


def jlog(logger, typ, item):
    func = getattr(logger, typ)
    func(json.dumps(item, indent=2, sort_keys=True))


def resp2json(resp):
    return {'message': resp.message, 'status': resp.status,
            'headers': resp.headers}


from otest.operation import Operation


class Done(Operation):
    def run(self, *args, **kwargs):
        self.conv.events.store(EV_END,'')
